from collections import deque
from http import HTTPStatus
from statistics import median
from typing import Any
import pytest
from faust import Event, Stream, Table, Topic
from faust.transport.consumer import Consumer
from faust.transport.producer import Producer
from faust.types import Message, TP
from faust.sensors.monitor import Monitor, TableState
from mode.utils.mocks import AsyncMock, Mock

TP1 = TP('foo', 0)


class test_Monitor:

    @pytest.fixture
    def time(self):
        timefun = Mock(name='time()')
        timefun.return_value = 101.1
        return timefun

    @pytest.fixture
    def message(self):
        return Mock(name='message', autospec=Message)

    @pytest.fixture
    def stream(self):
        return Mock(name='stream', autospec=Stream)

    @pytest.fixture
    def topic(self):
        return Mock(name='topic', autospec=Topic)

    @pytest.fixture
    def event(self):
        return Mock(name='event', autospec=Event)

    @pytest.fixture
    def table(self):
        return Mock(name='table', autospec=Table)

    @pytest.fixture
    def mon(self, *, time):
        mon = self.create_monitor()
        mon.time = time
        return mon

    def create_monitor(self, **kwargs: Any) -> Monitor:
        return Monitor(**kwargs)

    def create_populated_monitor(self,
            messages_active=101,
            messages_received_total=1001,
            messages_sent=303,
            messages_s=1000,
            messages_received_by_topic={'foo': 103},  # noqa
            events_active=202,
            events_total=2002,
            events_s=3000,
            events_runtime_avg=0.03,
            events_by_task={'mytask': 105},  # noqa
            events_by_stream={'stream': 105},  # noqa
            commit_latency=[1.03, 2.33, 16.33],  # noqa
            send_latency=[0.01, 0.04, 0.06, 0.010],  # noqa
            topic_buffer_full={'topic': 808},  # noqa
            **kwargs):
        return self.create_monitor(
            messages_active=messages_active,
            messages_received_total=messages_received_total,
            messages_sent=messages_sent,
            messages_s=messages_s,
            messages_received_by_topic=messages_received_by_topic,
            events_active=events_active,
            events_total=events_total,
            events_s=events_s,
            events_runtime_avg=events_runtime_avg,
            events_by_task=events_by_task,
            events_by_stream=events_by_stream,
            commit_latency=commit_latency,
            send_latency=send_latency,
            topic_buffer_full=topic_buffer_full,
            **kwargs)

    def test_init_max_avg_history(self):
        assert Monitor().max_avg_history == Monitor.max_avg_history

    def test_init_max_avg_history__default(self):
        assert Monitor(max_avg_history=33).max_avg_history == 33

    def test_init_max_commit_latency_history(self):
        assert (Monitor().max_commit_latency_history ==
                Monitor.max_commit_latency_history)

    def test_init_max_commit_latency_history__default(self):
        assert Monitor(
            max_commit_latency_history=33,
        ).max_commit_latency_history == 33

    def test_init_max_send_latency_history(self):
        assert (Monitor().max_send_latency_history ==
                Monitor.max_send_latency_history)

    def test_init_max_send_latency_history__default(self):
        assert Monitor(
            max_send_latency_history=33,
        ).max_send_latency_history == 33

    def test_init_max_assignment_latency_history(self):
        assert (Monitor().max_assignment_latency_history ==
                Monitor.max_assignment_latency_history)

    def test_init_max_assignment_latency_history__default(self):
        assert Monitor(
            max_assignment_latency_history=33,
        ).max_assignment_latency_history == 33

    def test_init_rebalances(self):
        assert Monitor(rebalances=99).rebalances == 99

    def test_asdict(self):
        mon = self.create_populated_monitor()
        assert mon.asdict() == {
            'messages_active': mon.messages_active,
            'messages_received_total': mon.messages_received_total,
            'messages_sent': mon.messages_sent,
            'messages_sent_by_topic': mon.messages_sent_by_topic,
            'messages_s': mon.messages_s,
            'messages_received_by_topic': mon.messages_received_by_topic,
            'events_active': mon.events_active,
            'events_total': mon.events_total,
            'events_s': mon.events_s,
            'events_runtime_avg': mon.events_runtime_avg,
            'events_by_task': mon._events_by_task_dict(),
            'events_by_stream': mon._events_by_stream_dict(),
            'commit_latency': mon.commit_latency,
            'send_latency': mon.send_latency,
            'assignment_latency': mon.assignment_latency,
            'assignments_completed': mon.assignments_completed,
            'assignments_failed': mon.assignments_failed,
            'send_errors': mon.send_errors,
            'topic_buffer_full': mon._topic_buffer_full_dict(),
            'metric_counts': mon._metric_counts_dict(),
            'tables': {
                name: table.asdict() for name, table in mon.tables.items()
            },
            'topic_committed_offsets': {},
            'topic_read_offsets': {},
            'topic_end_offsets': {},
            'rebalance_end_avg': mon.rebalance_end_avg,
            'rebalance_end_latency': mon.rebalance_end_latency,
            'rebalance_return_avg': mon.rebalance_return_avg,
            'rebalance_return_latency': mon.rebalance_return_latency,
            'rebalances': mon.rebalances,
            'http_response_codes': mon._http_response_codes_dict(),
            'http_response_latency': mon.http_response_latency,
            'http_response_latency_avg': mon.http_response_latency_avg,
        }

    def test_on_message_in(self, *, message, mon, time):
        for i in range(1, 11):
            offset = 3 + i
            mon.on_message_in(TP1, offset, message)

            assert mon.messages_received_total == i
            assert mon.messages_active == i
            assert mon.messages_received_by_topic[TP1.topic] == i
            assert message.time_in is time()
            assert mon.tp_read_offsets[TP1] == offset

    def test_on_stream_event_in(self, *, event, mon, stream, time):
        for i in range(1, 11):
            state = mon.on_stream_event_in(TP1, 3 + i, stream, event)

            assert mon.events_total == i
            assert mon.events_by_stream[str(stream)] == i
            assert mon.events_by_task[str(stream.task_owner)] == i
            assert mon.events_active == i
            assert state == {
                'time_in': time(),
                'time_out': None,
                'time_total': None,
            }

    def test_on_stream_event_out(self, *, event, mon, stream, time):
        other_time = 303.3
        mon.events_active = 10
        for i in range(1, 11):
            state = {
                'time_in': other_time,
                'time_out': None,
                'time_total': None,
            }
            mon.on_stream_event_out(TP1, 3 + i, stream, event, state)

            assert mon.events_active == 10 - i
            assert state == {
                'time_in': other_time,
                'time_out': time(),
                'time_total': time() - other_time,
            }
            assert mon.events_runtime[-1] == time() - other_time

    def test_on_stream_event_out__missing_state(
            self, *, event, mon, stream, time):
        # should not be an error
        mon.on_stream_event_out(TP1, 3, stream, event, None)

    def test_on_topic_buffer_full(self, *, mon, topic):
        for i in range(1, 11):
            mon.on_topic_buffer_full(topic)

            assert mon.topic_buffer_full[topic] == i

    def test_on_message_out(self, *, message, mon, time):
        mon.messages_active = 10
        message.time_in = 10.7

        for i in range(1, 11):
            mon.on_message_out(TP1, 3 + i, message)

            assert mon.messages_active == 10 - i
            assert message.time_out == time()
            assert message.time_total == time() - message.time_in
        message.time_in = None
        mon.on_message_out(TP1, 3 + 11, message)

    def test_on_table_get(self, *, mon, table):
        for i in range(1, 11):
            mon.on_table_get(table, 'k')
            assert mon._table_or_create(table).keys_retrieved == i

    def test_on_table_set(self, *, mon, table):
        for i in range(1, 11):
            mon.on_table_set(table, 'k', 'v')
            assert mon._table_or_create(table).keys_updated == i

    def test_on_table_del(self, *, mon, table):
        for i in range(1, 11):
            mon.on_table_del(table, 'k')
            assert mon._table_or_create(table).keys_deleted == i

    def test_on_commit_initiated(self, *, mon, time):
        assert mon.on_commit_initiated(
            Mock(name='consumer', autospec=Consumer)) == time()

    def test_on_commit_completed(self, *, mon, time):
        other_time = 56.7
        mon.on_commit_completed(
            Mock(name='consumer', autospec=Consumer), other_time)
        assert mon.commit_latency[-1] == time() - other_time

    def test_on_send_initiated(self, *, mon, time):
        for i in range(1, 11):
            state = mon.on_send_initiated(
                Mock(name='producer', autospec=Producer),
                'topic', 'message', 2, 4)
            assert mon.messages_sent == i
            assert mon.messages_sent_by_topic['topic'] == i
            assert state == time()

    def test_on_send_completed(self, *, mon, time):
        other_time = 56.7
        mon.on_send_completed(
            Mock(name='producer', autospec=Producer),
            other_time,
            Mock(name='metadata'),
        )
        assert mon.send_latency[-1] == time() - other_time

    def test_on_send_error(self, *, mon, time):
        mon.on_send_error(
            Mock(name='producer', autospec=Producer),
            Mock(name='state'),
            KeyError('foo'),
        )
        assert mon.send_errors == 1

    def test_on_assignment_start(self, *, mon, time):
        state = mon.on_assignment_start(Mock(name='assignor'))
        assert state['time_start'] == time()

    def test_on_assignment_completed(self, *, mon, time):
        other_time = 56.7
        assignor = Mock(name='assignor')
        assert mon.assignments_completed == 0
        mon.on_assignment_completed(assignor, {'time_start': other_time})
        assert mon.assignment_latency[-1] == time() - other_time
        assert mon.assignments_completed == 1

    def test_on_assignment_error(self, *, mon, time):
        other_time = 56.7
        assignor = Mock(name='assignor')
        assert mon.assignments_failed == 0
        mon.on_assignment_error(
            assignor, {'time_start': other_time}, KeyError())
        assert mon.assignment_latency[-1] == time() - other_time
        assert mon.assignments_failed == 1

    def test_on_rebalance_start(self, *, mon, time, app):
        assert mon.rebalances == 0
        app.rebalancing_count = 1
        state = mon.on_rebalance_start(app)
        assert state['time_start'] == time()
        assert mon.rebalances == 1

    def test_on_rebalance_return(self, *, mon, time, app):
        other_time = 56.7
        state = {'time_start': other_time}
        mon.on_rebalance_return(app, state)
        assert mon.rebalance_return_latency[-1] == time() - other_time
        assert state['time_return'] == time()
        assert state['latency_return'] == time() - other_time

    def test_on_rebalance_end(self, *, mon, time, app):
        other_time = 56.7
        state = {'time_start': other_time}
        mon.on_rebalance_end(app, state)
        assert mon.rebalance_end_latency[-1] == time() - other_time
        assert state['time_end'] == time()
        assert state['latency_end'] == time() - other_time

    def test_on_web_request_start(self, *, mon, time, app):
        request = Mock(name='request')
        view = Mock(name='view')
        state = mon.on_web_request_start(app, request, view=view)
        assert state['time_start'] == time()

    def test_on_web_request_end(self, *, mon, time, app):
        response = Mock(name='response')
        response.status = 404
        self.assert_on_web_request_end(mon, time, app, response,
                                       expected_status=404)

    def test_on_web_request_end__None_response(self, *, mon, time, app):
        self.assert_on_web_request_end(mon, time, app, None,
                                       expected_status=500)

    def assert_on_web_request_end(self, mon, time, app, response,
                                  expected_status):
        request = Mock(name='request')
        view = Mock(name='view')
        other_time = 156.9
        state = {'time_start': other_time}
        mon.on_web_request_end(app, request, response, state, view=view)
        assert state['time_end'] == time()
        assert state['latency_end'] == time() - other_time
        assert state['status_code'] == HTTPStatus(expected_status)

        assert mon.http_response_latency[-1] == time() - other_time
        assert mon.http_response_codes[HTTPStatus(expected_status)] == 1

    def test_TableState_asdict(self, *, mon, table):
        state = mon._table_or_create(table)
        assert isinstance(state, TableState)
        assert state.table is table
        assert state.keys_retrieved == 0
        assert state.keys_updated == 0
        assert state.keys_deleted == 0

        expected_asdict = {
            'keys_retrieved': 0,
            'keys_updated': 0,
            'keys_deleted': 0,
        }
        assert state.asdict() == expected_asdict
        assert state.__reduce_keywords__() == {
            **state.asdict(),
            'table': table,
        }

    def test_on_tp_commit(self, *, mon):
        topic = 'foo'
        for offset in range(20):
            partitions = list(range(4))
            tps = {TP(topic=topic, partition=p) for p in partitions}
            commit_offsets = {tp: offset for tp in tps}
            mon.on_tp_commit(commit_offsets)
            assert all(mon.tp_committed_offsets[tp] == commit_offsets[tp]
                       for tp in tps)
            offsets_dict = mon.asdict()['topic_committed_offsets'][topic]
            assert all(offsets_dict[p] == offset for p in partitions)

    def test_track_tp_end_offsets(self, *, mon):
        tp = TP(topic='foo', partition=2)
        for offset in range(20):
            mon.track_tp_end_offset(tp, offset)
            assert mon.tp_end_offsets[tp] == offset
            offsets_dict = mon.asdict()['topic_end_offsets'][tp.topic]
            assert offsets_dict[tp.partition] == offset

    @pytest.mark.asyncio
    async def test_service_sampler(self, *, mon):
        mon = Monitor()

        i = 0
        mon.events_runtime = []
        mon.sleep = AsyncMock(name='sleep')

        def on_sample(prev_events, prev_messages):
            nonlocal i
            mon.events_runtime.append(i + 0.34)
            i += 1
            if i > 10:
                mon._stopped.set()
            return prev_events, prev_messages
        mon._sample = Mock(name='_sample')
        mon._sample.side_effect = on_sample

        await mon._sampler(mon)

    def test__sample(self, *, mon):
        prev_event_total = 0
        prev_message_total = 0
        mon.events_runtime = []
        mon._sample(prev_event_total, prev_message_total)
        mon.events_runtime = deque(range(100))
        mon.rebalance_return_latency = deque(range(100))
        mon.rebalance_end_latency = deque(range(100))
        mon.http_response_latency = deque(range(100))
        prev_event_total = 0
        prev_message_total = 0
        mon._sample(prev_event_total, prev_message_total)

        assert mon.events_runtime_avg == median(mon.events_runtime)
        assert mon.events_s == 0  # XXX this is wrong!

        assert mon.rebalance_return_avg == median(mon.rebalance_return_latency)
        assert mon.rebalance_end_avg == median(mon.rebalance_end_latency)
        assert mon.http_response_latency_avg == median(
            mon.http_response_latency)



"""LiveCheck - Test runner."""
import asyncio
import logging
import traceback
import typing

from time import monotonic
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
)

from mode.utils.logging import CompositeLogger
from mode.utils.times import humanize_seconds
from mode.utils.typing import NoReturn

from faust.models import maybe_model

from .exceptions import (
    LiveCheckError,
    TestFailed,
    TestRaised,
    TestSkipped,
    TestTimeout,
)
from .locals import current_test_stack
from .models import State, TestExecution, TestReport
from .signals import BaseSignal

if typing.TYPE_CHECKING:
    from .case import Case as _Case
else:
    class _Case: ...  # noqa

__all__ = ['TestRunner']


class TestRunner:
    """Execute and keep track of test instance."""

    case: _Case

    state: State = State.INIT
    test: TestExecution

    started: float
    ended: Optional[float]
    runtime: Optional[float]

    logs: List[Tuple[str, Tuple]]
    signal_latency: Dict[str, float]

    report: Optional[TestReport] = None
    error: Optional[BaseException] = None

    def __init__(self,
                 case: _Case,
                 test: TestExecution,
                 started: float) -> None:
        self.case = case
        self.test = test
        self.started = started
        self.ended = None
        self.runtime = None
        self.logs = []
        self.log = CompositeLogger(
            self.case.log.logger,
            formatter=self._format_log,
        )
        self.signal_latency = {}

    async def execute(self) -> None:
        """Execute this test."""
        test = self.test
        with current_test_stack.push(test):
            if not self.case.active:
                await self.skip('case inactive')
            elif test.is_expired:
                await self.skip('expired')
            args = self._prepare_args(test.test_args)
            kwargs = self._prepare_kwargs(test.test_kwargs)
            await self.on_start()
            try:
                await self.case.run(*args, **kwargs)
            except asyncio.CancelledError:
                pass
            except TestSkipped as exc:
                await self.on_skipped(exc)
                raise
            except TestTimeout as exc:
                await self.on_timeout(exc)
                raise
            except AssertionError as exc:
                await self.on_failed(exc)
                raise TestFailed(exc) from exc
            except LiveCheckError as exc:
                await self.on_error(exc)
                raise
            except Exception as exc:
                await self.on_error(exc)
                raise TestRaised(exc) from exc
            else:
                await self.on_pass()

    async def skip(self, reason: str) -> NoReturn:
        """Skip this test execution."""
        exc = TestSkipped(f'Test {self.test.ident} skipped: {reason}')
        try:
            raise exc
        except TestSkipped as exc:
            # save with traceback
            await self.on_skipped(exc)
            raise
        else:  # pragma: no cover
            assert False  # can not get here

    def _prepare_args(self, args: Iterable) -> Tuple:
        to_value = self._prepare_val
        return tuple(to_value(arg) for arg in args)

    def _prepare_kwargs(self, kwargs: Mapping[str, Any]) -> Mapping[str, Any]:
        to_value = self._prepare_val
        return {
            to_value(k): to_value(v)
            for k, v in kwargs.items()
        }

    def _prepare_val(self, arg: Any) -> Any:
        return maybe_model(arg)

    def _format_log(self, severity: int, msg: str,
                    *args: Any, **kwargs: Any) -> str:
        return f'[{self.test.shortident}] {msg}'

    async def on_skipped(self, exc: TestSkipped) -> None:
        """Call when a test execution was skipped."""
        self.state = State.SKIP
        self.log.info('Skipped expired test: %s expires=%s',
                      self.test.ident, self.test.expires)
        await self.case.on_test_skipped(self)

    async def on_start(self) -> None:
        """Call when a test starts executing."""
        self.log_info('≈≈≈ Test %s executing... (issued %s) ≈≈≈',
                      self.case.name, self.test.human_date)
        await self.case.on_test_start(self)

    async def on_signal_wait(self, signal: BaseSignal, timeout: float) -> None:
        """Call when the test is waiting for a signal."""
        self.log_info('∆ %r/%r %s (%rs)...',
                      signal.index,
                      self.case.total_signals,
                      signal.name.upper(),
                      timeout)

    async def on_signal_received(self,
                                 signal: BaseSignal,
                                 time_start: float,
                                 time_end: float) -> None:
        """Call when a signal related to this test is received."""
        latency = time_end - time_start
        self.signal_latency[signal.name] = latency

    async def on_failed(self, exc: BaseException) -> None:
        """Call when an invariant in the test has failed."""
        self.end()
        self.error = exc
        self.state = State.FAIL
        self.log.exception('Test failed: %r', exc)
        await self.case.on_test_failed(self, exc)
        await self._finalize_report()

    async def on_error(self, exc: BaseException) -> None:
        """Call when test execution raises error."""
        self.end()
        self.state = State.ERROR
        self.error = exc
        self.log.exception('Test raised: %r', exc)
        await self.case.on_test_error(self, exc)
        await self._finalize_report()

    async def on_timeout(self, exc: BaseException) -> None:
        """Call when test execution times out."""
        self.end()
        self.error = exc
        self.state = State.TIMEOUT
        self.log.exception('Test timed-out: %r', exc)
        await self.case.on_test_timeout(self, exc)
        await self._finalize_report()

    async def on_pass(self) -> None:
        """Call when test execution returns successfully."""
        self.end()
        self.error = None
        self.state = State.PASS
        human_secs = humanize_seconds(
            self.runtime or 0.0,
            microseconds=True,
            now='~0.0 seconds',
        )
        self.log_info('Test OK in %s √', human_secs)
        self._flush_logs()
        await self.case.on_test_pass(self)
        await self._finalize_report()

    async def _finalize_report(self) -> None:
        tb: Optional[str] = None
        error = self.error
        if error:
            tb = '\n'.join(traceback.format_tb(error.__traceback__))
        self.report = TestReport(
            case_name=self.case.name,
            state=self.state,
            test=self.test,
            runtime=self.runtime,
            signal_latency=self.signal_latency,
            error=str(error) if error else None,
            traceback=tb,
        )
        await self.case.post_report(self.report)

    def log_info(self, msg: str, *args: Any) -> None:
        """Log information related to the current execution."""
        if self.case.realtime_logs:
            self.log.info(msg, *args)
        else:
            self.logs.append((msg, args))

    def end(self) -> None:
        """End test execution."""
        self.ended = monotonic()
        self.runtime = self.ended - self.started

    def _flush_logs(self, severity: int = logging.INFO) -> None:
        logs = self.logs
        try:
            self.log.logger.log(severity, '\n'.join(
                self._format_log(severity, msg % log_args)  # noqa: S001
                for msg, log_args in logs))
        finally:
            logs.clear()



"""LiveCheck Signals - Test communication and synchronization."""
import asyncio
import typing

from time import monotonic
from typing import Any, Dict, Generic, Tuple, Type, TypeVar, cast
from mode import Seconds, want_seconds

from faust.models import maybe_model

from .exceptions import TestTimeout
from .locals import current_test_stack
from .models import SignalEvent

if typing.TYPE_CHECKING:
    from .case import Case as _Case
else:
    class _Case: ...  # noqa

__all__ = ['BaseSignal', 'Signal']

VT = TypeVar('VT')


class BaseSignal(Generic[VT]):
    """Generic base class for signals."""

    name: str
    case: _Case
    index: int

    def __init__(self,
                 name: str = '',
                 case: _Case = None,
                 index: int = -1) -> None:
        self.name = name
        self.case = cast(_Case, case)
        self.index = index

    async def send(self, value: VT = None, *,
                   key: Any = None,
                   force: bool = False) -> None:
        """Notify test that this signal is now complete."""
        raise NotImplementedError()

    async def wait(self, *,
                   key: Any = None,
                   timeout: Seconds = None) -> VT:
        """Wait for signal to be completed."""
        raise NotImplementedError()

    async def resolve(self, key: Any, event: SignalEvent) -> None:
        """Resolve signal with value."""
        self._set_current_value(key, event)
        self._wakeup_resolvers()

    def __set_name__(self, owner: Type, name: str) -> None:
        if not self.name:
            self.name = name

    def _wakeup_resolvers(self) -> None:
        self.case.app._can_resolve.set()

    async def _wait_for_resolved(self, *, timeout: float = None) -> None:
        app = self.case.app
        app._can_resolve.clear()
        await app.wait(app._can_resolve, timeout=timeout)

    def _get_current_value(self, key: Any) -> SignalEvent:
        return self.case.app._resolved_signals[self._index_key(key)]

    def _index_key(self, key: Any) -> Tuple[str, str, Any]:
        return self.name, self.case.name, key

    def _set_current_value(self, key: Any, event: SignalEvent) -> None:
        self.case.app._resolved_signals[self._index_key(key)] = event

    def clone(self, **kwargs: Any) -> 'BaseSignal':
        """Clone this signal using keyword arguments."""
        return type(self)(**{**self._asdict(), **kwargs})

    def _asdict(self, **kwargs: Any) -> Dict:
        return {'name': self.name, 'case': self.case, 'index': self.index}

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.name}>'


class Signal(BaseSignal[VT]):
    """Signal for test case using Kafka.

    Used to wait for something to happen elsewhere.

    """

    # What do we use for this? Kafka? some other mechanism?
    # I'm thinking separate Kafka cluster, with a single
    # topic for each test app.

    async def send(self, value: VT = None, *,
                   key: Any = None,
                   force: bool = False) -> None:
        """Notify test that this signal is now complete."""
        current_test = current_test_stack.top
        if current_test is None:
            if not force:
                return
            assert key
        else:
            key = key if key is not None else current_test.id
        await self.case.app.bus.send(
            key=key,
            value=SignalEvent(
                signal_name=self.name,
                case_name=self.case.name,
                key=key,
                value=value,
            ),
        )

    async def wait(self, *,
                   key: Any = None,
                   timeout: Seconds = None) -> VT:
        """Wait for signal to be completed."""
        # wait for key to arrive in consumer
        runner = self.case.current_execution
        if runner is None:
            raise RuntimeError('No test executing.')
        test = runner.test
        assert test
        k: Any = test.id if key is None else key
        timeout_s = want_seconds(timeout)
        await runner.on_signal_wait(self, timeout=timeout_s)
        time_start = monotonic()
        event = await self._wait_for_message_by_key(key=k, timeout=timeout_s)
        time_end = monotonic()
        await runner.on_signal_received(
            self,
            time_start=time_start,
            time_end=time_end,
        )
        self._verify_event(event, k, self.name, self.case.name)
        return cast(VT, maybe_model(event.value))

    def _verify_event(self,
                      ev: SignalEvent,
                      key: Any,
                      name: str,
                      case: str) -> None:
        assert ev.key == key, f'{ev.key!r} == {key!r}'
        assert ev.signal_name == name, f'{ev.signal_name!r} == {name!r}'
        assert ev.case_name == case, f'{ev.case_name!r} == {case!r}'

    async def _wait_for_message_by_key(
            self, key: Any, *,
            timeout: float = None,
            max_interval: float = 2.0) -> SignalEvent:
        app = self.case.app
        time_start = monotonic()
        remaining = timeout
        # See if the key is already there.
        try:
            return self._get_current_value(key)
        except KeyError:
            pass
        # If not, wait for it to arrive.
        while not app.should_stop:
            if remaining is not None:
                remaining = remaining - (monotonic() - time_start)
            try:
                if remaining is not None and remaining <= 0.0:
                    try:
                        return self._get_current_value(key)
                    except KeyError:
                        raise asyncio.TimeoutError() from None
                max_wait = None
                if remaining is not None:
                    max_wait = min(remaining, max_interval)
                await self._wait_for_resolved(timeout=max_wait)
            except asyncio.TimeoutError:
                msg = f'Timed out waiting for signal {self.name} ({timeout})'
                raise TestTimeout(msg) from None
            if app.should_stop:
                break
            try:
                val = self._get_current_value(key)
                return val
            except KeyError:
                pass
        raise asyncio.CancelledError()



import pytest
from mode.utils.mocks import AsyncMock, Mock
from faust.web import Request, View, Web
from faust.web.exceptions import MethodNotAllowed


@View.from_handler
async def foo(self, request):
    return self, request


class test_View:

    @pytest.fixture
    def web(self):
        return Mock(name='web', autospec=Web)

    @pytest.fixture
    def view(self, *, app, web):
        return foo(app, web)

    def test_from_handler(self):
        assert issubclass(foo, View)

    def test_from_handler__not_callable(self):
        with pytest.raises(TypeError):
            View.from_handler(1)

    def test_init(self, *, app, web, view):
        assert view.app is app
        assert view.web is web
        assert view.methods == {
            'head': view.head,
            'get': view.get,
            'post': view.post,
            'patch': view.patch,
            'delete': view.delete,
            'put': view.put,
            'options': view.options,
            'search': view.search,
        }

    @pytest.mark.parametrize('method', [
        'GET', 'POST', 'PATCH', 'DELETE', 'PUT', 'OPTIONS', 'SEARCH',
        'get', 'post', 'patch', 'delete', 'put', 'options', 'search',
    ])
    @pytest.mark.asyncio
    async def test_dispatch(self, method, *, view):
        request = Mock(name='request', autospec=Request)
        request.method = method
        request.match_info = {}
        handler = AsyncMock(name=method)
        view.methods[method.lower()] = handler
        assert await view(request) is handler.coro()
        handler.assert_called_once_with(request)

    def test_path_for(self, *, view):
        ret = view.path_for('name', foo=1)
        assert ret is view.web.url_for.return_value
        view.web.url_for.assert_called_once_with('name', foo=1)

    def test_url_for__no_base(self, *, view):
        view.app.conf.canonical_url = 'http://example.com/'
        ret = view.url_for('name', foo=1)
        assert ret

    def test_url_for__base(self, *, view):
        ret = view.url_for('name', 'http://example.bar', foo=1)
        assert ret

    @pytest.mark.asyncio
    async def test_get(self, *, view):
        req = Mock(name='request', autospec=Request)
        assert (await view.methods['get'](req)) == (view, req)

    @pytest.mark.asyncio
    async def test_interface_get(self, *, app, web):
        view = View(app, web)
        with pytest.raises(MethodNotAllowed):
            await view.get(Mock(name='request', autospec=Request))

    @pytest.mark.asyncio
    async def test_interface_post(self, *, view):
        with pytest.raises(MethodNotAllowed):
            await view.post(Mock(name='request', autospec=Request))

    @pytest.mark.asyncio
    async def test_interface_put(self, *, view):
        with pytest.raises(MethodNotAllowed):
            await view.put(Mock(name='request', autospec=Request))

    @pytest.mark.asyncio
    async def test_interface_patch(self, *, view):
        with pytest.raises(MethodNotAllowed):
            await view.patch(Mock(name='request', autospec=Request))

    @pytest.mark.asyncio
    async def test_interface_delete(self, *, view):
        with pytest.raises(MethodNotAllowed):
            await view.delete(Mock(name='request', autospec=Request))

    @pytest.mark.asyncio
    async def test_interface_options(self, *, view):
        with pytest.raises(MethodNotAllowed):
            await view.options(Mock(name='request', autospec=Request))

    @pytest.mark.asyncio
    async def test_interface_search(self, *, view):
        with pytest.raises(MethodNotAllowed):
            await view.search(Mock(name='request', autospec=Request))

    def test_text(self, *, view, web):
        response = view.text(
            'foo',
            content_type='app/json',
            status=101,
            reason='foo',
            headers={'k': 'v'},
        )
        web.text.assert_called_once_with(
            'foo',
            content_type='app/json',
            status=101,
            reason='foo',
            headers={'k': 'v'},
        )
        assert response is web.text()

    def test_html(self, *, view, web):
        response = view.html(
            'foo',
            status=101,
            content_type='text/html',
            reason='bar',
            headers={'k': 'v'},
        )
        web.html.assert_called_once_with(
            'foo',
            status=101,
            content_type='text/html',
            reason='bar',
            headers={'k': 'v'},
        )
        assert response is web.html()

    def test_json(self, *, view, web):
        response = view.json(
            'foo',
            status=101,
            content_type='application/json',
            reason='bar',
            headers={'k': 'v'},
        )
        web.json.assert_called_once_with(
            'foo',
            status=101,
            content_type='application/json',
            reason='bar',
            headers={'k': 'v'},
        )
        assert response is web.json()

    def test_bytes(self, *, view, web):
        response = view.bytes(
            'foo',
            content_type='app/json',
            status=101,
            reason='foo',
            headers={'k': 'v'},
        )
        web.bytes.assert_called_once_with(
            'foo',
            content_type='app/json',
            status=101,
            reason='foo',
            headers={'k': 'v'},
        )
        assert response is web.bytes()

    def test_route(self, *, view, web):
        handler = Mock(name='handler')
        res = view.route('pat', handler)
        web.route.assert_called_with('pat', handler)
        assert res is handler

    def test_error(self, *, view, web):
        response = view.error(303, 'foo', arg='bharg')
        web.json.assert_called_once_with(
            {'error': 'foo', 'arg': 'bharg'},
            status=303,
            reason=None,
            headers=None,
            content_type=None,
        )
        assert response is web.json()

    def test_notfound(self, *, view, web):
        response = view.notfound(arg='bharg')
        web.json.assert_called_once_with(
            {'error': 'Not Found', 'arg': 'bharg'},
            status=404,
            reason=None,
            headers=None,
            content_type=None,
        )
        assert response is web.json()

    @pytest.mark.asyncio
    async def test_read_request_content(self, *, view, web):
        request = Mock(name='request')
        web.read_request_content = AsyncMock()
        ret = await view.read_request_content(request)
        web.read_request_content.assert_called_once_with(request)
        assert ret is web.read_request_content.coro.return_value



"""Base class for table storage drivers."""
import abc

from collections.abc import ItemsView, KeysView, ValuesView
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from mode import Service
from yarl import URL

from faust.types import (
    AppT,
    CodecArg,
    CollectionT,
    EventT,
    ModelArg,
    StoreT,
    TP,
)
from faust.types.stores import KT, VT

__all__ = ['Store', 'SerializedStore']


class Store(StoreT[KT, VT], Service):
    """Base class for table storage drivers."""

    def __init__(self,
                 url: Union[str, URL],
                 app: AppT,
                 table: CollectionT,
                 *,
                 table_name: str = '',
                 key_type: ModelArg = None,
                 value_type: ModelArg = None,
                 key_serializer: CodecArg = None,
                 value_serializer: CodecArg = None,
                 options: Mapping[str, Any] = None,
                 **kwargs: Any) -> None:
        Service.__init__(self, **kwargs)
        self.url = URL(url)
        self.app = app
        self.table = table
        self.table_name = table_name or self.table.name
        self.key_type = key_type
        self.value_type = value_type
        self.key_serializer = key_serializer
        self.value_serializer = value_serializer
        self.options = options

    def __hash__(self) -> int:
        return object.__hash__(self)

    def persisted_offset(self, tp: TP) -> Optional[int]:
        """Return the persisted offset for this topic and partition."""
        raise NotImplementedError('In-memory store only, does not persist.')

    def set_persisted_offset(self, tp: TP, offset: int) -> None:
        """Set the persisted offset for this topic and partition."""
        ...

    async def need_active_standby_for(self, tp: TP) -> bool:
        """Return :const:`True` if we have a copy of standby from elsewhere."""
        return True

    async def on_rebalance(self,
                           table: CollectionT,
                           assigned: Set[TP],
                           revoked: Set[TP],
                           newly_assigned: Set[TP]) -> None:
        """Handle rebalancing of the cluster."""
        ...

    async def on_recovery_completed(self,
                                    active_tps: Set[TP],
                                    standby_tps: Set[TP]) -> None:
        """Signal that table recovery completed."""
        ...

    def _encode_key(self, key: KT) -> bytes:
        key_bytes = self.app.serializers.dumps_key(
            self.key_type, key, serializer=self.key_serializer)
        if key_bytes is None:
            raise TypeError('Table key cannot be None')
        return key_bytes

    def _encode_value(self, value: VT) -> Optional[bytes]:
        return self.app.serializers.dumps_value(
            self.value_type, value, serializer=self.value_serializer)

    def _decode_key(self, key: Optional[bytes]) -> KT:
        return cast(KT, self.app.serializers.loads_key(
            self.key_type, key, serializer=self.key_serializer))

    def _decode_value(self, value: Optional[bytes]) -> VT:
        return self.app.serializers.loads_value(
            self.value_type, value, serializer=self.value_serializer)

    def _repr_info(self) -> str:
        return f'table_name={self.table_name} url={self.url}'

    def prefix_scan(self, prefix: bytes) -> Iterator[Tuple[bytes, bytes]]:
        ...

    @property
    def label(self) -> str:
        """Return short description of this store."""
        return f'{type(self).__name__}: {self.url}'


class _SerializedStoreKeysView(KeysView):

    def __init__(self, store: 'SerializedStore') -> None:
        self._mapping = store

    def __iter__(self) -> Iterator:
        yield from self._mapping._keys_decoded()


class _SerializedStoreValuesView(ValuesView):

    def __init__(self, store: 'SerializedStore') -> None:
        self._mapping = store

    def __iter__(self) -> Iterator:
        yield from self._mapping._values_decoded()


class _SerializedStoreItemsView(ItemsView):

    def __init__(self, store: 'SerializedStore') -> None:
        self._mapping = store

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        yield from self._mapping._items_decoded()


class SerializedStore(Store[KT, VT]):
    """Base class for table storage drivers requiring serialization."""

    @abc.abstractmethod
    def _get(self, key: bytes, partition: Any = None) -> Optional[bytes]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _set(self,
             key: bytes,
             value: Optional[bytes],
             partition: Any = None) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _del(self, key: bytes, partition: Any = None) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _iterkeys(self) -> Iterator[bytes]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _itervalues(self) -> Iterator[bytes]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _iteritems(self) -> Iterator[Tuple[bytes, bytes]]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _size(self) -> int:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _contains(self, key: bytes) -> bool:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _clear(self) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    def _prefix_scan(self, prefix: bytes) -> Iterator[Tuple[bytes, bytes]]:  # pragma: no cover
        ...

    def apply_changelog_batch(self, batch: Iterable[EventT],
                              to_key: Callable[[Any], KT],
                              to_value: Callable[[Any], VT]) -> None:
        """Apply batch of events from changelog topic to this store."""
        for event in batch:
            key = event.message.key
            if key is None:
                raise TypeError(
                    f'Changelog entry is missing key: {event.message}')
            value = event.message.value
            if value is None:
                self._del(key)
            else:
                # keys/values are already JSON serialized in the message
                self._set(key, value)

    def __getitem__(self, key: KT) -> VT:
        value = self._get(self._encode_key(key))
        if value is None:
            raise KeyError(key)
        return self._decode_value(value)

    def __setitem__(self, key: KT, value: VT) -> None:
        return self._set(self._encode_key(key), self._encode_value(value))

    def __delitem__(self, key: KT) -> None:
        return self._del(self._encode_key(key))

    def __iter__(self) -> Iterator[KT]:
        yield from self._keys_decoded()

    def __len__(self) -> int:
        return self._size()

    def __contains__(self, key: object) -> bool:
        return self._contains(self._encode_key(cast(KT, key)))

    def keys(self) -> KeysView:
        """Return view of keys in the K/V store."""
        return _SerializedStoreKeysView(self)

    def _keys_decoded(self) -> Iterator[KT]:
        for key in self._iterkeys():
            yield self._decode_key(key)

    def values(self) -> ValuesView:
        """Return view of values in the K/V store."""
        return _SerializedStoreValuesView(self)

    def _values_decoded(self) -> Iterator[VT]:
        for value in self._itervalues():
            yield self._decode_value(value)

    def items(self) -> ItemsView:
        """Return view of items in the K/V store as (key, value) pairs."""
        return _SerializedStoreItemsView(self)

    def _items_decoded(self) -> Iterator[Tuple[KT, VT]]:
        for key, value in self._iteritems():
            yield self._decode_key(key), self._decode_value(value)

    def clear(self) -> None:
        """Clear all data from this K/V store."""
        self._clear()

    def update(self, *args: Any, partition: int = None, **kwargs: Any) -> None:
        for d in args:
            for key, value in d.items():
                self._set(self._encode_key(key), self._encode_value(value), partition=partition)
        for key, value in kwargs.items():
            self._set(self._encode_key(key), self._encode_value(value), partition=partition)

    def get_for_partition(self, key: KT, partition: int) -> VT:
        value = self._get(self._encode_key(key), partition=partition)
        return self._decode_value(value) if value is not None else None
    
    def del_for_partition(self, key, partition: int) -> None:
        return self._del(self._encode_key(key), partition=partition)
    
    def prefix_scan(self, prefix: KT) -> Iterator[Tuple[KT, VT]]:
        # json serializer wraps prefix value in quotes
        # we truncate the last quote character to ensure the prefix is matched
        if self.key_serializer == 'json':
            _prefix = self._encode_key(prefix)[:-1]
        else:
            _prefix = self._encode_key(prefix)
        for key, value in self._prefix_scan(_prefix):
            yield self._decode_key(key), self._decode_value(value)    


"""Program ``faust worker`` used to start application from console."""
from typing import Any
from .worker import worker

__all__ = ['livecheck']


class livecheck(worker):
    """Manage LiveCheck instances."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        try:
            livecheck = self.app.livecheck  # type: ignore
        except AttributeError:
            raise self.UsageError(
                f'App {self.app!r} is not associated with LiveCheck')
        self.app = livecheck
        self._finalize_concrete_app(self.app)



"""Admin related Kafka protocol extensions."""
from aiokafka.protocol import types
from .api import Request, Response


class CreateTopicsResponse_v0(Response):
    """Response from Create Topic request (version 0)."""

    API_KEY = 19
    API_VERSION = 0
    SCHEMA = types.Schema(
        ('topic_error_codes', types.Array(
            ('topic', types.String('utf-8')),
            ('error_code', types.Int16))),
    )


class CreateTopicsResponse_v1(Response):
    """Response from Create Topic request (version 1)."""

    API_KEY = 19
    API_VERSION = 1
    SCHEMA = types.Schema(
        ('topic_error_codes', types.Array(
            ('topic', types.String('utf-8')),
            ('error_code', types.Int16),
            ('error_message', types.String('utf-8')))),
    )


class CreateTopicsRequest_v0(Request):
    """Request to create topic (version 0)."""

    API_KEY = 19
    API_VERSION = 0
    RESPONSE_TYPE = CreateTopicsResponse_v0
    SCHEMA = types.Schema(
        ('create_topic_requests', types.Array(
            ('topic', types.String('utf-8')),
            ('num_partitions', types.Int32),
            ('replication_factor', types.Int16),
            ('replica_assignment', types.Array(
                ('partition_id', types.Int32),
                ('replicas', types.Array(types.Int32)))),
            ('configs', types.Array(
                ('config_key', types.String('utf-8')),
                ('config_value', types.String('utf-8')))))),
        ('timeout', types.Int32),
    )


class CreateTopicsRequest_v1(Request):
    """Request to create topic (version 1)."""

    API_KEY = 19
    API_VERSION = 1
    RESPONSE_TYPE = CreateTopicsResponse_v1
    SCHEMA = types.Schema(
        ('create_topic_requests', types.Array(
            ('topic', types.String('utf-8')),
            ('num_partitions', types.Int32),
            ('replication_factor', types.Int16),
            ('replica_assignment', types.Array(
                ('partition_id', types.Int32),
                ('replicas', types.Array(types.Int32)))),
            ('configs', types.Array(
                ('config_key', types.String('utf-8')),
                ('config_value', types.String('utf-8')))))),
        ('timeout', types.Int32),
        ('validate_only', types.Boolean),
    )


CreateTopicsRequest = [CreateTopicsRequest_v0, CreateTopicsRequest_v1]
CreateTopicsResponse = [CreateTopicsResponse_v0, CreateTopicsRequest_v1]



import ssl
from enum import Enum
from typing import Optional, Union

__all__ = [
    'AUTH_PROTOCOLS_SSL',
    'AUTH_PROTOCOLS_SASL',
    'AuthProtocol',
    'CredentialsArg',
    'CredentialsT',
    'SASLMechanism',
    'to_credentials',
]


class AuthProtocol(Enum):
    SSL = 'SSL'
    PLAINTEXT = 'PLAINTEXT'
    SASL_PLAINTEXT = 'SASL_PLAINTEXT'
    SASL_SSL = 'SASL_SSL'


class SASLMechanism(Enum):
    PLAIN = 'PLAIN'
    GSSAPI = 'GSSAPI'
    SCRAM_SHA_256 = 'SCRAM-SHA-256'
    SCRAM_SHA_512 = 'SCRAM-SHA-512'


AUTH_PROTOCOLS_SSL = {AuthProtocol.SSL, AuthProtocol.SASL_SSL}
AUTH_PROTOCOLS_SASL = {AuthProtocol.SASL_PLAINTEXT, AuthProtocol.SASL_SSL}


class CredentialsT:
    protocol: AuthProtocol


CredentialsArg = Union[CredentialsT, ssl.SSLContext]


def to_credentials(obj: CredentialsArg = None) -> Optional[CredentialsT]:
    if obj is not None:
        from faust.auth import SSLCredentials  # XXX :(
        if isinstance(obj, ssl.SSLContext):
            return SSLCredentials(obj)
        if isinstance(obj, CredentialsT):
            return obj
        from faust.exceptions import ImproperlyConfigured
        raise ImproperlyConfigured(
            f'Unknown credentials type {type(obj)}: {obj}')
    return None



"""Partition assignor."""
import socket
import zlib
import time

from collections import defaultdict
from typing import (
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Set,
    cast,
)
from aiokafka.cluster import ClusterMetadata
from aiokafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from aiokafka.coordinator.protocol import (
    ConsumerProtocolMemberAssignment,
    ConsumerProtocolMemberMetadata,
)
from mode import get_logger
from yarl import URL

from faust.types.app import AppT
from faust.types.assignor import (
    HostToPartitionMap,
    PartitionAssignorT,
    TopicToPartitionMap,
)
from faust.types.tables import TableManagerT
from faust.types.tuples import TP

from .client_assignment import ClientAssignment, ClientMetadata
from .cluster_assignment import ClusterAssignment
from .copartitioned_assignor import CopartitionedAssignor

__all__ = [
    'MemberAssignmentMapping',
    'MemberMetadataMapping',
    'MemberSubscriptionMapping',
    'ClientMetadataMapping',
    'ClientAssignmentMapping',
    'CopartitionedGroups',
    'PartitionAssignor',
]

MemberAssignmentMapping = MutableMapping[str, ConsumerProtocolMemberAssignment]
MemberMetadataMapping = MutableMapping[str, ConsumerProtocolMemberMetadata]
MemberSubscriptionMapping = MutableMapping[str, List[str]]
ClientMetadataMapping = MutableMapping[str, ClientMetadata]
ClientAssignmentMapping = MutableMapping[str, ClientAssignment]
CopartitionedGroups = MutableMapping[int, Iterable[Set[str]]]

logger = get_logger(__name__)


class PartitionAssignor(
        AbstractPartitionAssignor, PartitionAssignorT):  # type: ignore
    """PartitionAssignor handles internal topic creation.

    Further, this assignor needs to be sticky and potentially redundant

    Notes:
        Interface copied from :mod:`kafka.coordinator.assignors.abstract`.
    """

    _assignment: ClientAssignment
    _table_manager: TableManagerT
    _member_urls: MutableMapping[str, str]
    _changelog_distribution: HostToPartitionMap
    _active_tps: Set[TP]
    _standby_tps: Set[TP]
    _tps_url: MutableMapping[TP, str]
    _topic_groups: MutableMapping[str, int]

    def __init__(self, app: AppT, replicas: int = 0) -> None:
        AbstractPartitionAssignor.__init__(self)
        self.app = app
        self._table_manager = self.app.tables
        self._assignment = ClientAssignment(actives={}, standbys={})
        self._changelog_distribution = {}
        self.replicas = replicas
        self._member_urls = {}
        self._tps_url = {}
        self._active_tps = set()
        self._standby_tps = set()
        self._topic_groups = {}
        self._assignment_retry_count = 0
        self._max_assignment_retries = 30
        self._last_assignment_attempt = None

    def group_for_topic(self, topic: str) -> int:
        return self._topic_groups[topic]

    @property
    def changelog_distribution(self) -> HostToPartitionMap:
        return self._changelog_distribution

    @changelog_distribution.setter
    def changelog_distribution(self, value: HostToPartitionMap) -> None:
        self._changelog_distribution = value
        self._tps_url = {
            TP(topic, partition): url
            for url, tps in self._changelog_distribution.items()
            for topic, partitions in tps.items() for partition in partitions
        }

    @property
    def _metadata(self) -> ClientMetadata:
        return ClientMetadata(
            assignment=self._assignment,
            url=str(self._url),
            changelog_distribution=self.changelog_distribution,
            topic_groups=self._topic_groups,
        )

    @property
    def _url(self) -> URL:
        return self.app.conf.canonical_url

    def on_assignment(
            self, assignment: ConsumerProtocolMemberMetadata) -> None:
        """Handle assignment from Kafka coordinator with retry logic."""
        metadata = cast(ClientMetadata,
                        ClientMetadata.loads(
                            self._decompress(assignment.user_data)))
        
        # Track assignment attempts
        current_time = time.time()
        if (self._last_assignment_attempt is None or 
            current_time - self._last_assignment_attempt > 30):  # Reset after 30s
            self._assignment_retry_count = 0
        self._last_assignment_attempt = current_time

        self._assignment = metadata.assignment
        self._topic_groups = dict(metadata.topic_groups)
        self._active_tps = self._assignment.active_tps
        self._standby_tps = self._assignment.standby_tps
        self.changelog_distribution = metadata.changelog_distribution
        
        # Verify assignment consistency
        kafka_assignment = sorted(assignment.assignment)
        faust_assignment = sorted(
            self._assignment.kafka_protocol_assignment(self._table_manager))
        
        if kafka_assignment != faust_assignment:
            self._assignment_retry_count += 1

            if self._assignment_retry_count <= self._max_assignment_retries:
                # Log and retry
                logger.warning(
                    'Assignment mismatch (attempt %d/%d) - requesting rebalance',
                    self._assignment_retry_count, self._max_assignment_retries
                )
                # Request rebalance to recalculate assignments based on refreshed metadata
                self._request_rebalance_for_assignment_mismatch()
                return
            else:
                # Max retries exceeded - log error but continue
                logger.error(
                    'Assignment mismatch persists after %d retries. '
                    'Continuing with Kafka coordinator assignment. '
                    'This may indicate a bug or configuration issue.',
                    self._max_assignment_retries
                )
                self._log_assignment_mismatch(kafka_assignment, faust_assignment, metadata)
                # Continue with the assignment anyway                
        
        # Verify URL consistency
        if metadata.url != str(self._url):
            logger.warning(
                'URL mismatch: expected=%r, got=%r - requesting rebalance',
                str(self._url), metadata.url
            )
            if self._assignment_retry_count <= self._max_assignment_retries:
                self._assignment_retry_count += 1
                self._request_rebalance_for_assignment_mismatch()
                return
        
        # Reset retry count on successful assignment
        self._assignment_retry_count = 0
        logger.debug('Assignment verification successful')

    def metadata(self, topics: Set[str]) -> ConsumerProtocolMemberMetadata:
        return ConsumerProtocolMemberMetadata(self.version, list(topics),
                                              self._metadata.dumps())

    def _request_rebalance_for_assignment_mismatch(self):
        """Request a rebalance when assignment mismatch is detected."""
        logger.info('Requesting rebalance due to assignment mismatch')
        # This will cause all members to rejoin and recalculate assignments
        self.app.consumer.request_rejoin()

    @classmethod
    def _group_co_subscribed(cls, topics: Set[str],
                             subscriptions: MemberSubscriptionMapping,
                             ) -> Iterable[Set[str]]:
        topic_subscriptions: MutableMapping[str, Set[str]] = defaultdict(set)
        for client, subscription in subscriptions.items():
            for topic in subscription:
                topic_subscriptions[topic].add(client)
        co_subscribed: MutableMapping[Sequence[str], Set[str]] = defaultdict(
            set)
        for topic in topics:
            clients = topic_subscriptions[topic]
            assert clients, 'Subscribed clients for topic cannot be empty'
            co_subscribed[tuple(clients)].add(topic)
        return co_subscribed.values()

    @classmethod
    def _get_copartitioned_groups(
            cls, topics: Set[str],
            cluster: ClusterMetadata,
            subscriptions: MemberSubscriptionMapping) -> CopartitionedGroups:
        topics_by_partitions: MutableMapping[int, Set] = defaultdict(set)
        for topic in topics:
            num_partitions = len(cluster.partitions_for_topic(topic) or set())
            if num_partitions == 0:
                logger.warning('Ignoring missing topic: %r', topic)
                continue
            topics_by_partitions[num_partitions].add(topic)
        # We group copartitioned topics by subscribed clients such that
        # a group of co-subscribed topics with the same number of partitions
        # are copartitioned
        copart_grouped = {
            num_partitions: cls._group_co_subscribed(topics, subscriptions)
            for num_partitions, topics in topics_by_partitions.items()
        }
        return copart_grouped

    @classmethod
    def _get_client_metadata(
            cls, metadata: ConsumerProtocolMemberMetadata) -> ClientMetadata:
        client_metadata = ClientMetadata.loads(metadata.user_data)
        return cast(ClientMetadata, client_metadata)

    def _update_member_urls(self,
                            clients_metadata: ClientMetadataMapping) -> None:
        self._member_urls = {
            member_id: client_metadata.url
            for member_id, client_metadata in clients_metadata.items()
        }

    def assign(
            self,
            cluster: ClusterMetadata,
            member_metadata: MemberMetadataMapping) -> MemberAssignmentMapping:
        if self.app.tracer:
            return self._trace_assign(cluster, member_metadata)
        else:
            return self._assign(cluster, member_metadata)

    def _trace_assign(
            self,
            cluster: ClusterMetadata,
            member_metadata: MemberMetadataMapping) -> MemberAssignmentMapping:
        assert self.app.tracer is not None
        span = self.app.tracer.get_tracer('_faust').start_span(
            operation_name='coordinator_assignment',
            tags={'hostname': socket.gethostname()},
        )
        with span:
            assignment = self._assign(cluster, member_metadata)
            self.app._span_add_default_tags(span)
            span.set_tag('assignment', assignment)
        return assignment

    def _assign(
            self,
            cluster: ClusterMetadata,
            member_metadata: MemberMetadataMapping) -> MemberAssignmentMapping:
        sensor_state = self.app.sensors.on_assignment_start(self)
        try:
            assignment = self._perform_assignment(cluster, member_metadata)
        except MemoryError:
            raise
        except Exception as exc:
            self.app.sensors.on_assignment_error(self, sensor_state, exc)
        else:
            self.app.sensors.on_assignment_completed(self, sensor_state)
        return assignment

    def _perform_assignment(
            self,
            cluster: ClusterMetadata,
            member_metadata: MemberMetadataMapping) -> MemberAssignmentMapping:
        cluster_assgn = ClusterAssignment()

        clients_metadata = {
            member_id: self._get_client_metadata(metadata)
            for member_id, metadata in member_metadata.items()
        }

        subscriptions = {
            member_id: cast(List[str], metadata.subscription)
            for member_id, metadata in member_metadata.items()
        }

        for member_id in member_metadata:
            cluster_assgn.add_client(member_id, subscriptions[member_id],
                                     clients_metadata[member_id])
        topics = cluster_assgn.topics()

        copartitioned_groups = self._get_copartitioned_groups(
            topics, cluster, subscriptions)

        self._update_member_urls(clients_metadata)

        # Initialize fresh assignment
        assignments: ClientAssignmentMapping = {
            member_id: ClientAssignment(actives={}, standbys={})
            for member_id in member_metadata
        }

        topic_to_group_id = {}
        partitions_by_topic = {}

        for group_id, (num_partitions, topic_groups) in enumerate(sorted(
                copartitioned_groups.items())):
            for topics in topic_groups:
                for topic in topics:
                    topic_to_group_id[topic] = group_id
                    partitions_by_topic[topic] = num_partitions
                assert len(topics) > 0 and num_partitions > 0
                # Get assignment for unique copartitioned group
                assgn = cluster_assgn.copartitioned_assignments(topics)
                assignor = CopartitionedAssignor(
                    topics=topics,
                    cluster_asgn=assgn,
                    num_partitions=num_partitions,
                    replicas=self.replicas,
                )
                # Update client assignments for copartitioned group
                for client, copart_assn in assignor.get_assignment().items():
                    assignments[client].add_copartitioned_assignment(
                        copart_assn)

        # Add all changelogs of global tables as standby for all members
        assignments = self._global_table_standby_assignments(
            assignments, partitions_by_topic)

        changelog_distribution = self._get_changelog_distribution(assignments)
        res = self._protocol_assignments(
            assignments,
            changelog_distribution,
            topic_to_group_id,
        )
        return res

    def _global_table_standby_assignments(
            self,
            assignments: ClientAssignmentMapping,
            partitions_by_topic: Mapping[str, int]) -> ClientAssignmentMapping:
        # Ensures all members have access to all changelog partitions
        # as standbys, if not already as actives
        for table in self._table_manager.data.values():
            # Add changelog standbys only if global table
            if table.is_global:
                changelog_topic_name = table._changelog_topic_name()
                num_partitions = partitions_by_topic[changelog_topic_name]
                assert num_partitions is not None
                all_partitions = set(range(0, num_partitions))
                for assignment in assignments.values():
                    active_partitions = set(
                        assignment.actives.get(
                            changelog_topic_name, []))
                    # Only add those partitions as standby which aren't active
                    standby_partitions = all_partitions - active_partitions
                    assignment.standbys[
                        changelog_topic_name] = list(standby_partitions)
        return assignments

    def _protocol_assignments(
            self,
            assignments: ClientAssignmentMapping,
            cl_distribution: HostToPartitionMap,
            topic_groups: Mapping[str, int]) -> MemberAssignmentMapping:
        return {
            client: ConsumerProtocolMemberAssignment(
                self.version,
                sorted(
                    assignment.kafka_protocol_assignment(self._table_manager)),
                self._compress(
                    ClientMetadata(
                        assignment=assignment,
                        url=self._member_urls[client],
                        changelog_distribution=cl_distribution,
                        topic_groups=topic_groups,
                    ).dumps(),
                ),
            )
            for client, assignment in assignments.items()
        }

    @classmethod
    def _compress(cls, raw: bytes) -> bytes:
        return zlib.compress(raw)

    @classmethod
    def _decompress(cls, compressed: bytes) -> bytes:
        return zlib.decompress(compressed)

    @classmethod
    def _topics_filtered(cls, assignment: TopicToPartitionMap,
                         topics: Set[str]) -> TopicToPartitionMap:
        return {
            topic: partitions
            for topic, partitions in assignment.items() if topic in topics
        }

    def _get_changelog_distribution(
            self, assignments: ClientAssignmentMapping) -> HostToPartitionMap:
        topics = self._table_manager.changelog_topics
        return {
            self._member_urls[client]: self._topics_filtered(
                assignment.actives, topics)
            for client, assignment in assignments.items()
        }

    @property
    def name(self) -> str:
        return 'faust'

    @property
    def version(self) -> int:
        return 4

    def assigned_standbys(self) -> Set[TP]:
        return {
            TP(topic, partition)
            for topic, partitions in self._assignment.standbys.items()
            for partition in partitions
        }

    def assigned_actives(self) -> Set[TP]:
        return {
            TP(topic, partition)
            for topic, partitions in self._assignment.actives.items()
            for partition in partitions
        }

    def table_metadata(self, topic: str) -> HostToPartitionMap:
        return {
            host: self._topics_filtered(assignment, {topic})
            for host, assignment in self.changelog_distribution.items()
        }

    def tables_metadata(self) -> HostToPartitionMap:
        return self.changelog_distribution

    def key_store(self, topic: str, key: bytes) -> URL:
        return URL(self._tps_url[self.app.producer.key_partition(topic, key)])

    def is_active(self, tp: TP) -> bool:
        return tp in self._active_tps

    def is_standby(self, tp: TP) -> bool:
        return tp in self._standby_tps



from contextlib import ExitStack, contextmanager

import pytest
import terminaltables
from faust.utils.terminal import tables
from mode.utils.mocks import Mock, patch

TABLE_DATA = [
    ('foo', 1.0, 3.33, 6.66, 9.99),
    ('bar', 2.0, 2.34, 4.23, 3.33),
]


def fh(isatty=True):
    fh = Mock()
    fh.isatty.return_value = isatty
    return fh


@contextmanager
def mock_stdout(isatty=True):
    with patch('sys.stdout') as stdout:
        stdout.isatty.return_value = isatty
        yield


@contextmanager
def mock_logging(isatty=True):
    with patch('mode.utils.logging.LOG_ISATTY', isatty):
        yield


@pytest.mark.parametrize('target,contexts,kwargs,expected_tty', [
    (None, [mock_stdout(isatty=True)], {}, True),
    (None, [mock_stdout(isatty=True)], {'kw': 1}, True),
    (None, [mock_stdout(isatty=False)], {}, False),
    (None, [mock_stdout(isatty=False)], {'kw': 2}, False),
    (fh(isatty=True), [], {}, True),
    (fh(isatty=True), [], {'kw': 1}, True),
    (fh(isatty=False), [], {}, False),
    (fh(isatty=False), [], {'kw': 1}, False),
])
def test_table(target, contexts, kwargs, expected_tty):
    with ExitStack() as stack:
        for context in contexts:
            stack.enter_context(context)
        with patch('faust.utils.terminal.tables._get_best_table_type') as _g:
            table = tables.table(
                TABLE_DATA, title='Title', target=target, **kwargs)
            _g.assert_called_with(expected_tty)
            _g.return_value.assert_called_with(
                TABLE_DATA, title='Title', **kwargs)
            assert table is _g.return_value.return_value


@pytest.mark.parametrize('tty,contexts,headers,expected_tty,expected_data', [
    (None, [mock_logging(isatty=True)], None, True, TABLE_DATA),
    (None, [mock_logging(isatty=True)], ['foo'], True, [['foo']] + TABLE_DATA),
    (None, [mock_logging(isatty=False)], None, False, TABLE_DATA),
    (None, [mock_logging(isatty=False)], ['f'], False, [['f']] + TABLE_DATA),
    (True, [], None, True, TABLE_DATA),
    (True, [], ['foo'], True, [['foo']] + TABLE_DATA),
    (False, [], None, False, TABLE_DATA),
    (False, [], ['foo'], False, [['foo']] + TABLE_DATA),
])
def test_logtable(tty, contexts, headers, expected_tty, expected_data):
    with ExitStack() as stack:
        for context in contexts:
            stack.enter_context(context)
        with patch('faust.utils.terminal.tables.table') as table:
            ret = tables.logtable(
                TABLE_DATA,
                title='Title',
                target=None,
                tty=tty,
                headers=headers)
            table.assert_called_with(
                expected_data, title='Title', target=None, tty=expected_tty,
            )
            assert ret is table().table


@pytest.mark.parametrize('tty,expected_table_type', [
    (True, terminaltables.SingleTable),
    (False, terminaltables.AsciiTable),
])
def test_get_best_table_type(tty, expected_table_type):
    assert tables._get_best_table_type(tty) is expected_table_type


def test_table__default_tty():
    with patch('faust.utils.terminal.tables._get_best_table_type') as g:
        with patch('faust.utils.terminal.tables.isatty') as i:
            i.return_value = None
            tables.table({}, tty=None, title='foo')
            g.assert_called_once_with(False)



"""Models used by agents internally."""
from typing import Any

from faust.models import Record
from faust.types import ModelT

__all__ = ['ReqRepRequest', 'ReqRepResponse']


class ReqRepRequest(Record,
                    serializer='json',
                    namespace='@ReqRepRequest',  # internal namespace
                    # any stream should allow this type
                    # to wrap other values.
                    polymorphic_fields=True):
    """Value wrapped in a Request-Reply request."""

    # agent.ask(value) wraps the value in this record
    # so that the receiving agent knows where to send the reply.

    value: Any
    reply_to: str
    correlation_id: str


class ModelReqRepRequest(ReqRepRequest):
    """Request-Reply request where value is a model."""

    value: ModelT


class ReqRepResponse(Record, serializer='json', namespace='@ReqRepResponse'):
    """Request-Reply response."""

    key: Any
    value: Any
    correlation_id: str


class ModelReqRepResponse(ReqRepResponse):
    value: Any



import abc
from typing import Any, Optional, Tuple, Union

__all__ = ['CodecArg', 'CodecT']


class CodecT(metaclass=abc.ABCMeta):
    """Abstract type for an encoder/decoder.

    See Also:
        :class:`faust.serializers.codecs.Codec`.
    """

    @abc.abstractmethod
    def __init__(self, children: Tuple['CodecT', ...] = None,
                 **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    def dumps(self, obj: Any) -> bytes:
        ...

    @abc.abstractmethod
    def loads(self, s: bytes) -> Any:
        ...

    @abc.abstractmethod
    def clone(self, *children: 'CodecT') -> 'CodecT':
        ...

    @abc.abstractmethod
    def __or__(self, other: Any) -> Any:
        ...


# `serializer` argument can be str or Codec instance.
CodecArg = Optional[Union[CodecT, str]]



import pytest
from yarl import URL
from faust.utils.urls import urllist


def test_urllist_URL():
    assert urllist(URL('foo://localhost')) == [URL('foo://localhost')]


@pytest.mark.parametrize('value', [
    None,
    '',
])
def test_urllist_empty_raises(value):
    with pytest.raises(ValueError):
        urllist(value)


def test_urllist_str():
    assert urllist('foo://localhost') == [URL('foo://localhost')]


def test_urllist_str_no_scheme():
    assert urllist('bar.com', default_scheme='foo') == [URL('foo://bar.com')]


def test_urllist_URL_no_scheme_scalar_arg():
    assert urllist(URL('bar.com'), default_scheme='foo') == [
        URL('foo://bar.com'),
    ]


def test_urllist_URL_no_scheme_list_arg():
    assert urllist([URL('bar.com'), URL('//localhost:9092')],
                   default_scheme='foo') == [
        URL('foo://bar.com'),
        URL('foo://localhost:9092'),
    ]


def test_urllist_strsep():
    assert urllist('foo://localhost;bar.com;example.com') == [
        URL('foo://localhost'),
        URL('foo://bar.com'),
        URL('foo://example.com'),
    ]


def test_urllist_strsep_no_scheme():
    assert urllist('localhost;bar.com;example.com', default_scheme='bar') == [
        URL('bar://localhost'),
        URL('bar://bar.com'),
        URL('bar://example.com'),
    ]


def test_urllist_list_of_strings():
    assert urllist(['kafka://kafka1.example.com:9092',
                    'kafka://kafka2.example.com:9092',
                    ]) == [URL('kafka://kafka1.example.com:9092'),
                           URL('kafka://kafka2.example.com:9092')]


def test_urllist_URLs():
    assert urllist([
        URL('foo://localhost'),
        URL('bar.com'),
        URL('example.com'),
    ]) == [
        URL('foo://localhost'),
        URL('foo://bar.com'),
        URL('foo://example.com'),
    ]


def test_urllist_URLs_no_scheme():
    assert urllist([
        URL('localhost'),
        URL('bar.com'),
        URL('example.com'),
    ], default_scheme='foo') == [
        URL('foo://localhost'),
        URL('foo://bar.com'),
        URL('foo://example.com'),
    ]



import aiokafka
import faust
import opentracing
import pytest
import random
import string
from contextlib import contextmanager
from typing import Optional

from aiokafka.errors import CommitFailedError, IllegalStateError, KafkaError
from aiokafka.structs import OffsetAndMetadata, TopicPartition
from opentracing.ext import tags
from faust import auth
from faust.exceptions import ImproperlyConfigured, NotReady
from faust.sensors.monitor import Monitor
from faust.transport.drivers import aiokafka as mod
from faust.transport.drivers.aiokafka import (
    AIOKafkaConsumerThread,
    Consumer,
    ConsumerNotStarted,
    ConsumerRebalanceListener,
    ConsumerStoppedError,
    Producer,
    ProducerSendError,
    TOPIC_LENGTH_MAX,
    Transport,
    credentials_to_aiokafka_auth,
    server_list,
)
from faust.types import TP
from mode.utils.futures import done_future
from mode.utils.mocks import ANY, AsyncMock, MagicMock, Mock, call, patch

TP1 = TP('topic', 23)
TP2 = TP('topix', 23)

TESTED_MODULE = 'faust.transport.drivers.aiokafka'


@pytest.fixture()
def thread():
    return Mock(
        name='thread',
        create_topic=AsyncMock(),
    )


@pytest.fixture()
def consumer(*,
             thread,
             app,
             callback,
             on_partitions_revoked,
             on_partitions_assigned):
    consumer = Consumer(
        app.transport,
        callback=callback,
        on_partitions_revoked=on_partitions_revoked,
        on_partitions_assigned=on_partitions_assigned,
    )
    consumer._thread = thread
    return consumer


@pytest.fixture()
def callback():
    return Mock(name='callback')


@pytest.fixture()
def on_partitions_revoked():
    return Mock(name='on_partitions_revoked')


@pytest.fixture()
def on_partitions_assigned():
    return Mock(name='on_partitions_assigned')


class test_ConsumerRebalanceListener:

    @pytest.fixture()
    def handler(self, *, thread):
        return ConsumerRebalanceListener(thread)

    @pytest.fixture()
    def thread(self):
        return Mock(
            name='thread',
            on_partitions_assigned=AsyncMock(),
            on_partitions_revoked=AsyncMock(),
        )

    @pytest.mark.asyncio
    async def test_on_partitions_revoked(self, *, handler, thread):
        await handler.on_partitions_revoked([
            TopicPartition('A', 0),
            TopicPartition('B', 3),
        ])
        thread.on_partitions_revoked.assert_called_once_with({
            TP('A', 0), TP('B', 3),
        })

    @pytest.mark.asyncio
    async def test_on_partitions_assigned(self, *, handler, thread):
        await handler.on_partitions_assigned([
            TopicPartition('A', 0),
            TopicPartition('B', 3),
        ])
        thread.on_partitions_assigned.assert_called_once_with({
            TP('A', 0), TP('B', 3),
        })


class test_Consumer:

    @pytest.fixture()
    def thread(self):
        return Mock(
            name='thread',
            create_topic=AsyncMock(),
        )

    @pytest.fixture()
    def consumer(self, *,
                 thread,
                 app,
                 callback,
                 on_partitions_revoked,
                 on_partitions_assigned):
        consumer = Consumer(
            app.transport,
            callback=callback,
            on_partitions_revoked=on_partitions_revoked,
            on_partitions_assigned=on_partitions_assigned,
        )
        consumer._thread = thread
        return consumer

    @pytest.fixture()
    def callback(self):
        return Mock(name='callback')

    @pytest.fixture()
    def on_partitions_revoked(self):
        return Mock(name='on_partitions_revoked')

    @pytest.fixture()
    def on_partitions_assigned(self):
        return Mock(name='on_partitions_assigned')

    @pytest.mark.asyncio
    async def test_create_topic(self, *, consumer, thread):
        await consumer.create_topic(
            'topic', 30, 3,
            timeout=40.0,
            retention=50.0,
            compacting=True,
            deleting=True,
            ensure_created=True,
        )
        thread.create_topic.assert_called_once_with(
            'topic',
            30,
            3,
            config=None,
            timeout=40.0,
            retention=50.0,
            compacting=True,
            deleting=True,
            ensure_created=True,
        )

    def test__new_topicpartition(self, *, consumer):
        tp = consumer._new_topicpartition('t', 3)
        assert isinstance(tp, TopicPartition)
        assert tp.topic == 't'
        assert tp.partition == 3

    def test__to_message(self, *, consumer):
        record = self.mock_record(
            timestamp=3000,
            headers=[('a', b'b')],
        )
        m = consumer._to_message(TopicPartition('t', 3), record)
        assert m.topic == record.topic
        assert m.partition == record.partition
        assert m.offset == record.offset
        assert m.timestamp == 3.0
        assert m.headers == record.headers
        assert m.key == record.key
        assert m.value == record.value
        assert m.checksum == record.checksum
        assert m.serialized_key_size == record.serialized_key_size
        assert m.serialized_value_size == record.serialized_value_size

    def test__to_message__no_timestamp(self, *, consumer):
        record = self.mock_record(timestamp=None)
        m = consumer._to_message(TopicPartition('t', 3), record)
        assert m.timestamp is None

    def mock_record(self,
                    topic='t',
                    partition=3,
                    offset=1001,
                    timestamp=None,
                    timestamp_type=1,
                    headers=None,
                    key=b'key',
                    value=b'value',
                    checksum=312,
                    serialized_key_size=12,
                    serialized_value_size=40,
                    **kwargs):
        return Mock(
            name='record',
            topic=topic,
            partition=partition,
            offset=offset,
            timestamp=timestamp,
            timestamp_type=timestamp_type,
            headers=headers,
            key=key,
            value=value,
            checksum=checksum,
            serialized_key_size=serialized_key_size,
            serialized_value_size=serialized_value_size,
        )

    @pytest.mark.asyncio
    async def test_on_stop(self, *, consumer):
        consumer.transport._topic_waiters = {'topic': Mock()}
        await consumer.on_stop()
        assert not consumer.transport._topic_waiters


class AIOKafkaConsumerThreadFixtures:

    @pytest.fixture()
    def cthread(self, *, consumer):
        return AIOKafkaConsumerThread(consumer)

    @pytest.fixture()
    def tracer(self, *, app):
        tracer = app.tracer = Mock(name='tracer')
        tobj = tracer.get_tracer.return_value

        def start_span(operation_name=None, **kwargs):
            span = opentracing.Span(
                tracer=tobj,
                context=opentracing.SpanContext(),
            )

            if operation_name is not None:
                span.operation_name = operation_name
                assert span.operation_name == operation_name
            return span

        tobj.start_span = start_span
        return tracer

    @pytest.fixture()
    def _consumer(self):
        return Mock(
            name='AIOKafkaConsumer',
            autospec=aiokafka.AIOKafkaConsumer,
            start=AsyncMock(),
            stop=AsyncMock(),
            commit=AsyncMock(),
            position=AsyncMock(),
            end_offsets=AsyncMock(),
        )

    @pytest.fixture()
    def now(self):
        return 1201230410

    @pytest.fixture()
    def tp(self):
        return TP('foo', 30)

    @pytest.fixture()
    def aiotp(self, *, tp):
        return TopicPartition(tp.topic, tp.partition)

    @pytest.fixture()
    def logger(self, *, cthread):
        cthread.log = Mock(name='cthread.log')
        return cthread.log


class test_verify_event_path_base(AIOKafkaConsumerThreadFixtures):

    last_request: Optional[float] = None
    last_response: Optional[float] = None
    highwater: int = 1
    committed_offset: int = 1
    acks_enabled: bool = False
    stream_inbound: Optional[float] = None
    last_commit: Optional[float] = None
    expected_message: Optional[str] = None
    has_monitor = True

    def _set_started(self, t):
        self._cthread.time_started = t

    def _set_last_request(self, last_request):
        self.__consumer.records_last_request[self._aiotp] = last_request

    def _set_last_response(self, last_response):
        self.__consumer.records_last_response[self._aiotp] = last_response

    def _set_stream_inbound(self, inbound_time):
        self._app.monitor.stream_inbound_time[self._tp] = inbound_time

    def _set_last_commit(self, commit_time):
        self._cthread.tp_last_committed_at[self._tp] = commit_time

    @pytest.fixture(autouse=True)
    def aaaa_setup_attributes(self, *,
                              app,
                              cthread,
                              _consumer,
                              now,
                              tp,
                              aiotp):
        self._app = app
        self._tp = tp
        self._aiotp = aiotp
        self._now = now
        self._cthread = cthread
        self.__consumer = _consumer

    @pytest.fixture(autouse=True)
    def setup_consumer(self, *, app, cthread, _consumer, now, tp, aiotp):
        assert self._tp is tp
        assert self._aiotp is aiotp
        # patch self.acks_enabledc
        app.topics.acks_enabled_for = Mock(name='acks_enabled_for')
        app.topics.acks_enabled_for.return_value = self.acks_enabled

        # patch consumer.time_started
        self._set_started(now)

        # connect underlying AIOKafkaConsumer object.
        cthread._consumer = _consumer

        # patch AIOKafkaConsumer.records_last_request to self.last_request
        _consumer.records_last_request = {}
        if self.last_request is not None:
            self._set_last_request(self.last_request)

        # patch AIOKafkaConsumer.records_last_response to self.last_response
        _consumer.records_last_response = {}
        if self.last_response is not None:
            self._set_last_response(self.last_response)

        # patch app.monitor
        if self.has_monitor:
            cthread.consumer.app.monitor = Mock(name='monitor', spec=Monitor)
            app.monitor = cthread.consumer.app.monitor
            app.monitor = Mock(name='monitor', spec=Monitor)
            # patch monitor.stream_inbound_time
            # this is the time when a stream last processed a record
            # for tp
            app.monitor.stream_inbound_time = {}
            self._set_stream_inbound(self.stream_inbound)
        else:
            app.monitor = None

        # patch highwater
        cthread.highwater = Mock(name='highwater')
        cthread.highwater.return_value = self.highwater

        # patch committed offset
        cthread.consumer._committed_offset = {
            tp: self.committed_offset,
        }

        cthread.tp_last_committed_at = {}
        self._set_last_commit(self.last_commit)

    def test_state(self, *, cthread, now):
        # verify that setup_consumer fixture was applied
        assert cthread.time_started == now


class test_VEP_no_fetch_since_start(test_verify_event_path_base):

    def test_just_started(self, *, cthread, now, tp, logger):
        self._set_started(now - 2.0)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out(self, *, cthread, now, tp, logger):
        self._set_started(
            now - cthread.tp_fetch_request_timeout_secs * 2,
        )
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_called_with(
            mod.SLOW_PROCESSING_NO_FETCH_SINCE_START,
            ANY, ANY,
        )


class test_VEP_no_response_since_start(test_verify_event_path_base):

    def test_just_started(self, *, cthread, _consumer, now, tp, logger):
        self._set_last_request(now - 5.0)
        self._set_started(now - 2.0)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out(self, *, cthread, _consumer, now, tp, logger):
        assert cthread.verify_event_path(now, tp) is None
        self._set_last_request(now - 5.0)
        self._set_started(
            now - cthread.tp_fetch_response_timeout_secs * 2,
        )
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_called_with(
            mod.SLOW_PROCESSING_NO_RESPONSE_SINCE_START,
            ANY, ANY,
        )


class test_VEP_no_recent_fetch(test_verify_event_path_base):

    def test_recent_fetch(self, *, cthread, now, tp, logger):
        self._set_last_response(now - 30.0)
        self._set_last_request(now - 2.0)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out(self, *, cthread, now, tp, logger):
        self._set_last_response(now - 30.0)
        self._set_last_request(now - cthread.tp_fetch_request_timeout_secs * 2)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_called_with(
            mod.SLOW_PROCESSING_NO_RECENT_FETCH,
            ANY, ANY,
        )


class test_VEP_no_recent_response(test_verify_event_path_base):

    def test_recent_response(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 2.0)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(
            now - cthread.tp_fetch_response_timeout_secs * 2)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_called_with(
            mod.SLOW_PROCESSING_NO_RECENT_RESPONSE,
            ANY, ANY,
        )


class test_VEP_no_highwater_since_start(test_verify_event_path_base):
    highwater = None

    def test_no_monitor(self, *, app, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now)
        app.monitor = None
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_just_started(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now - cthread.tp_stream_timeout_secs * 2)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_called_with(
            mod.SLOW_PROCESSING_NO_HIGHWATER_SINCE_START,
            ANY, ANY,
        )


class test_VEP_stream_idle_no_highwater(test_verify_event_path_base):

    highwater = 10
    committed_offset = 10

    def test_highwater_same_as_offset(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now - 300.0)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()


class test_VEP_stream_idle_highwater_no_acks(
        test_verify_event_path_base):
    acks_enabled = False

    def test_no_acks(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()


class test_VEP_stream_idle_highwater_same_has_acks_everything_OK(
        test_verify_event_path_base):
    highwater = 10
    committed_offset = 10
    inbound_time = None
    acks_enabled = True

    def test_main(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()


class test_VEP_stream_idle_highwater_no_inbound(
        test_verify_event_path_base):
    highwater = 20
    committed_offset = 10
    inbound_time = None
    acks_enabled = True

    def test_just_started(self, *, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out_since_start(self, *, app, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now - cthread.tp_stream_timeout_secs * 2)
        assert cthread.verify_event_path(now, tp) is None
        expected_message = cthread._make_slow_processing_error(
            mod.SLOW_PROCESSING_STREAM_IDLE_SINCE_START,
            [mod.SLOW_PROCESSING_CAUSE_STREAM,
             mod.SLOW_PROCESSING_CAUSE_AGENT])
        logger.error.assert_called_once_with(
            expected_message,
            tp, ANY,
            setting='stream_processing_timeout',
            current_value=app.conf.stream_processing_timeout,
        )

    def test_has_inbound(self, *, app, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now - cthread.tp_stream_timeout_secs * 2)
        self._set_stream_inbound(now)
        self._set_last_commit(now)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_inbound_timed_out(self, *, app, cthread, now, tp, logger):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now - cthread.tp_stream_timeout_secs * 4)
        self._set_stream_inbound(now - cthread.tp_stream_timeout_secs * 2)
        self._set_last_commit(now)
        assert cthread.verify_event_path(now, tp) is None
        expected_message = cthread._make_slow_processing_error(
            mod.SLOW_PROCESSING_STREAM_IDLE,
            [mod.SLOW_PROCESSING_CAUSE_STREAM,
             mod.SLOW_PROCESSING_CAUSE_AGENT])
        logger.error.assert_called_once_with(
            expected_message,
            tp, ANY,
            setting='stream_processing_timeout',
            current_value=app.conf.stream_processing_timeout,
        )


class test_VEP_no_commit(test_verify_event_path_base):
    highwater = 20
    committed_offset = 10
    inbound_time = None
    acks_enabled = True

    def _configure(self, now, cthread):
        self._set_last_request(now - 10.0)
        self._set_last_response(now - 5.0)
        self._set_started(now - cthread.tp_stream_timeout_secs * 4)
        self._set_stream_inbound(now - 0.01)

    def test_just_started(self, *, cthread, now, tp, logger):
        self._configure(now, cthread)
        self._set_last_commit(None)
        self._set_started(now)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()

    def test_timed_out_since_start(self, *, app, cthread, now, tp, logger):
        self._configure(now, cthread)
        self._set_last_commit(None)
        self._set_started(now - cthread.tp_commit_timeout_secs * 2)
        assert cthread.verify_event_path(now, tp) is None
        expected_message = cthread._make_slow_processing_error(
            mod.SLOW_PROCESSING_NO_COMMIT_SINCE_START,
            [mod.SLOW_PROCESSING_CAUSE_COMMIT],
        )
        logger.error.assert_called_once_with(
            expected_message,
            tp, ANY,
            setting='broker_commit_livelock_soft_timeout',
            current_value=app.conf.broker_commit_livelock_soft_timeout,
        )

    def test_timed_out_since_last(self, *, app, cthread, now, tp, logger):
        self._configure(now, cthread)
        self._set_last_commit(cthread.tp_commit_timeout_secs * 2)
        self._set_started(now - cthread.tp_commit_timeout_secs * 4)
        assert cthread.verify_event_path(now, tp) is None
        expected_message = cthread._make_slow_processing_error(
            mod.SLOW_PROCESSING_NO_RECENT_COMMIT,
            [mod.SLOW_PROCESSING_CAUSE_COMMIT],
        )
        logger.error.assert_called_once_with(
            expected_message,
            tp, ANY,
            setting='broker_commit_livelock_soft_timeout',
            current_value=app.conf.broker_commit_livelock_soft_timeout,
        )

    def test_committing_fine(self, *, app, cthread, now, tp, logger):
        self._configure(now, cthread)
        self._set_last_commit(now - 2.0)
        self._set_started(now - cthread.tp_commit_timeout_secs * 4)
        assert cthread.verify_event_path(now, tp) is None
        logger.error.assert_not_called()


class test_AIOKafkaConsumerThread(AIOKafkaConsumerThreadFixtures):

    def test_constructor(self, *, cthread):
        assert cthread._partitioner
        assert cthread._rebalance_listener

    @pytest.mark.asyncio
    async def test_on_start(self, *, cthread, _consumer):
        cthread._create_consumer = Mock(
            name='_create_consumer',
            return_value=_consumer,
        )
        await cthread.on_start()

        assert cthread._consumer is cthread._create_consumer.return_value
        cthread._create_consumer.assert_called_once_with(
            loop=cthread.thread_loop)
        cthread._consumer.start.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_thread_stop(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        await cthread.on_thread_stop()
        cthread._consumer.stop.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_thread_stop__consumer_not_started(self, *, cthread):
        cthread._consumer = None
        await cthread.on_thread_stop()

    def test__create_consumer__client(self, *, cthread, app):
        app.client_only = True
        loop = Mock(name='loop')
        cthread._create_client_consumer = Mock(name='_create_client_consumer')
        c = cthread._create_consumer(loop=loop)
        assert c is cthread._create_client_consumer.return_value
        cthread._create_client_consumer.assert_called_once_with(
            cthread.transport, loop=loop)

    def test__create_consumer__worker(self, *, cthread, app):
        app.client_only = False
        loop = Mock(name='loop')
        cthread._create_worker_consumer = Mock(name='_create_worker_consumer')
        c = cthread._create_consumer(loop=loop)
        assert c is cthread._create_worker_consumer.return_value
        cthread._create_worker_consumer.assert_called_once_with(
            cthread.transport, loop=loop)

    def test_session_gt_request_timeout(self, *, cthread, app):
        app.conf.broker_session_timeout = 90
        app.conf.broker_request_timeout = 10

        with pytest.raises(ImproperlyConfigured):
            self.assert_create_worker_consumer(
                cthread, app,
                in_transaction=False)

    def test__create_worker_consumer(self, *, cthread, app):
        self.assert_create_worker_consumer(
            cthread, app,
            in_transaction=False,
            isolation_level='read_uncommitted',
        )

    def test__create_worker_consumer__transaction(self, *, cthread, app):
        self.assert_create_worker_consumer(
            cthread, app,
            in_transaction=True,
            isolation_level='read_committed',
        )

    def assert_create_worker_consumer(self, cthread, app,
                                      in_transaction=False,
                                      isolation_level='read_uncommitted',
                                      api_version=None):
        loop = Mock(name='loop')
        transport = cthread.transport
        conf = app.conf
        cthread.consumer.in_transaction = in_transaction
        auth_settings = credentials_to_aiokafka_auth(
            conf.broker_credentials, conf.ssl_context)
        with patch('aiokafka.AIOKafkaConsumer') as AIOKafkaConsumer:
            c = cthread._create_worker_consumer(transport, loop)
            assert c is AIOKafkaConsumer.return_value
            max_poll_interval = conf.broker_max_poll_interval
            AIOKafkaConsumer.assert_called_once_with(
                loop=loop,
                api_version=app.conf.consumer_api_version,
                client_id=conf.broker_client_id,
                group_id=conf.id,
                bootstrap_servers=server_list(
                    transport.url, transport.default_port),
                partition_assignment_strategy=[cthread._assignor],
                enable_auto_commit=False,
                auto_offset_reset=conf.consumer_auto_offset_reset,
                max_poll_records=conf.broker_max_poll_records,
                max_poll_interval_ms=int(max_poll_interval * 1000.0),
                max_partition_fetch_bytes=conf.consumer_max_fetch_size,
                fetch_max_wait_ms=1500,
                request_timeout_ms=int(conf.broker_request_timeout * 1000.0),
                rebalance_timeout_ms=int(
                    conf.broker_rebalance_timeout * 1000.0),
                check_crcs=conf.broker_check_crcs,
                session_timeout_ms=int(conf.broker_session_timeout * 1000.0),
                heartbeat_interval_ms=int(
                    conf.broker_heartbeat_interval * 1000.0),
                isolation_level=isolation_level,
                traced_from_parent_span=cthread.traced_from_parent_span,
                start_rebalancing_span=cthread.start_rebalancing_span,
                start_coordinator_span=cthread.start_coordinator_span,
                on_generation_id_known=cthread.on_generation_id_known,
                flush_spans=cthread.flush_spans,
                **auth_settings,
            )

    def test__create_client_consumer(self, *, cthread, app):
        loop = Mock(name='loop')
        transport = cthread.transport
        conf = app.conf
        auth_settings = credentials_to_aiokafka_auth(
            conf.broker_credentials, conf.ssl_context)
        with patch('aiokafka.AIOKafkaConsumer') as AIOKafkaConsumer:
            c = cthread._create_client_consumer(transport, loop)
            max_poll_interval = conf.broker_max_poll_interval
            assert c is AIOKafkaConsumer.return_value
            AIOKafkaConsumer.assert_called_once_with(
                loop=loop,
                client_id=conf.broker_client_id,
                bootstrap_servers=server_list(
                    transport.url, transport.default_port),
                request_timeout_ms=int(conf.broker_request_timeout * 1000.0),
                max_poll_interval_ms=int(max_poll_interval * 1000.0),
                enable_auto_commit=True,
                max_poll_records=conf.broker_max_poll_records,
                auto_offset_reset=conf.consumer_auto_offset_reset,
                check_crcs=conf.broker_check_crcs,
                **auth_settings,
            )

    def test__start_span(self, *, cthread, app):
        with patch(TESTED_MODULE + '.set_current_span') as s:
            app.tracer = Mock(name='tracer')
            span = cthread._start_span('test')
            app.tracer.get_tracer.assert_called_once_with(
                f'{app.conf.name}-_aiokafka')
            tracer = app.tracer.get_tracer.return_value
            tracer.start_span.assert_called_once_with(
                operation_name='test')
            span.set_tag.assert_has_calls([
                call(tags.SAMPLING_PRIORITY, 1),
                call('faust_app', app.conf.name),
                call('faust_id', app.conf.id),
            ])
            s.assert_called_once_with(span)
            assert span is tracer.start_span.return_value

    def test_trace_category(self, *, cthread, app):
        assert cthread.trace_category == f'{app.conf.name}-_aiokafka'

    def test_transform_span_lazy(self, *, cthread, app, tracer):
        cthread._consumer = Mock(name='_consumer')
        cthread._consumer._coordinator.generation = -1
        self.assert_setup_lazy_spans(cthread, app, tracer)

        cthread._consumer._coordinator.generation = 10
        pending = cthread._pending_rebalancing_spans
        assert len(pending) == 3

        cthread.on_generation_id_known()
        assert not pending

    def test_transform_span_flush_spans(self, *, cthread, app, tracer):
        cthread._consumer = Mock(name='_consumer')
        cthread._consumer._coordinator.generation = -1
        self.assert_setup_lazy_spans(cthread, app, tracer)
        pending = cthread._pending_rebalancing_spans
        assert len(pending) == 3

        cthread.flush_spans()
        assert not pending

    def test_span_without_operation_name(self, *, cthread):
        span = opentracing.Span(
            tracer=Mock('tobj'),
            context=opentracing.SpanContext(),
        )

        assert cthread._on_span_cancelled_early(span) is None

    def test_transform_span_lazy_no_consumer(self, *, cthread, app, tracer):
        cthread._consumer = Mock(name='_consumer')
        cthread._consumer._coordinator.generation = -1
        self.assert_setup_lazy_spans(cthread, app, tracer)

        cthread._consumer = None
        pending = cthread._pending_rebalancing_spans
        assert len(pending) == 3

        while pending:
            span = pending.popleft()
            cthread._on_span_generation_known(span)

    def test_transform_span_eager(self, *, cthread, app, tracer):
        cthread._consumer = Mock(name='_consumer')
        cthread._consumer._coordinator.generation = 10
        self.assert_setup_lazy_spans(cthread, app, tracer, expect_lazy=False)

    def assert_setup_lazy_spans(self, cthread, app, tracer, expect_lazy=True):
        got_foo = got_bar = got_baz = False

        def foo():
            nonlocal got_foo
            got_foo = True
            T = cthread.traced_from_parent_span(None, lazy=True)
            T(bar)()

        def bar():
            nonlocal got_bar
            got_bar = True
            T = cthread.traced_from_parent_span(None, lazy=True)
            T(REPLACE_WITH_MEMBER_ID)()

        def REPLACE_WITH_MEMBER_ID():
            nonlocal got_baz
            got_baz = True

        with cthread.start_rebalancing_span() as span:
            T = cthread.traced_from_parent_span(span)
            T(foo)()
            if expect_lazy:
                assert len(cthread._pending_rebalancing_spans) == 2

        assert got_foo
        assert got_bar
        assert got_baz
        if expect_lazy:
            assert len(cthread._pending_rebalancing_spans) == 3
        else:
            assert not cthread._pending_rebalancing_spans

    def test__start_span__no_tracer(self, *, cthread, app):
        app.tracer = None
        with cthread._start_span('test') as span:
            assert span

    def test_traced_from_parent_span(self, *, cthread):
        with patch(TESTED_MODULE + '.traced_from_parent_span') as traced:
            parent_span = Mock(name='parent_span')
            ret = cthread.traced_from_parent_span(parent_span, foo=303)
            traced.assert_called_once_with(parent_span, callback=None, foo=303)
            assert ret is traced.return_value

    def test_start_rebalancing_span(self, *, cthread):
        cthread._start_span = Mock()
        ret = cthread.start_rebalancing_span()
        assert ret is cthread._start_span.return_value
        cthread._start_span.assert_called_once_with('rebalancing', lazy=True)

    def test_start_coordinator_span(self, *, cthread):
        cthread._start_span = Mock()
        ret = cthread.start_coordinator_span()
        assert ret is cthread._start_span.return_value
        cthread._start_span.assert_called_once_with('coordinator')

    def test_close(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread.close()
        _consumer.set_close.assert_called_once_with()
        _consumer._coordinator.set_close.assert_called_once_with()

    def test_close__no_consumer(self, *, cthread):
        cthread._consumer = None
        cthread.close()

    @pytest.mark.asyncio
    async def test_subscribe(self, *, cthread, _consumer):
        with self.assert_calls_thread(
                cthread, _consumer, _consumer.subscribe,
                topics={'foo', 'bar'},
                listener=cthread._rebalance_listener):
            await cthread.subscribe(['foo', 'bar'])

    @pytest.mark.asyncio
    async def test_seek_to_committed(self, *, cthread, _consumer):
        with self.assert_calls_thread(
                cthread, _consumer, _consumer.seek_to_committed):
            await cthread.seek_to_committed()

    @pytest.mark.asyncio
    async def test_commit(self, *, cthread, _consumer):
        offsets = {TP1: 100}
        with self.assert_calls_thread(
                cthread, _consumer, cthread._commit, offsets):
            await cthread.commit(offsets)

    @pytest.mark.asyncio
    async def test__commit(self, *, cthread, _consumer):
        offsets = {TP1: 1001}
        cthread._consumer = _consumer
        await cthread._commit(offsets)

        _consumer.commit.assert_called_once_with(
            {TP1: OffsetAndMetadata(1001, '')},
        )

    @pytest.mark.asyncio
    async def test__commit__already_rebalancing(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        _consumer.commit.side_effect = CommitFailedError('already rebalanced')
        assert not (await cthread._commit({TP1: 1001}))

    @pytest.mark.asyncio
    async def test__commit__CommitFailedError(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        exc = _consumer.commit.side_effect = CommitFailedError('xx')
        cthread.crash = AsyncMock()
        assert not (await cthread._commit({TP1: 1001}))
        cthread.crash.assert_called_once_with(exc)

    @pytest.mark.asyncio
    async def test__commit__IllegalStateError(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread.assignment = Mock()
        exc = _consumer.commit.side_effect = IllegalStateError('xx')
        cthread.crash = AsyncMock()
        assert not (await cthread._commit({TP1: 1001}))
        cthread.crash.assert_called_once_with(exc)

    @pytest.mark.asyncio
    async def test_position(self, *, cthread, _consumer):
        with self.assert_calls_thread(
                cthread, _consumer, _consumer.position, TP1):
            await cthread.position(TP1)

    @pytest.mark.asyncio
    async def test_seek_to_beginning(self, *, cthread, _consumer):
        partitions = (TP1,)
        with self.assert_calls_thread(
                cthread, _consumer, _consumer.seek_to_beginning,
                *partitions):
            await cthread.seek_to_beginning(*partitions)

    @pytest.mark.asyncio
    async def test_seek_wait(self, *, cthread, _consumer):
        partitions = {TP1: 1001}
        with self.assert_calls_thread(
                cthread, _consumer, cthread._seek_wait,
                _consumer, partitions):
            await cthread.seek_wait(partitions)

    @pytest.mark.asyncio
    async def test__seek_wait(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread.consumer._read_offset.clear()
        partitions = {TP1: 0, TP2: 3}

        await cthread._seek_wait(_consumer, partitions)

        assert cthread.consumer._read_offset[TP2] == 3
        assert TP1 not in cthread.consumer._read_offset

        _consumer.position.assert_has_calls([
            call(TP1),
            call(TP2),
        ])

    @pytest.mark.asyncio
    async def test__seek_wait__empty(self, *, cthread, _consumer):
        await cthread._seek_wait(_consumer, {})

    def test_seek(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread.seek(TP1, 10)
        _consumer.seek.assert_called_once_with(TP1, 10)

    def test_assignment(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        _consumer.assignment.return_value = {
            TopicPartition(TP1.topic, TP1.partition),
        }
        assignment = cthread.assignment()
        assert assignment == {TP1}
        assert all(isinstance(x, TP) for x in assignment)

    def test_highwater(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread.consumer.in_transaction = False
        ret = cthread.highwater(TP1)
        assert ret is _consumer.highwater.return_value
        _consumer.highwater.assert_called_once_with(TP1)

    def test_highwater__in_transaction(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread.consumer.in_transaction = True
        ret = cthread.highwater(TP1)
        assert ret is _consumer.last_stable_offset.return_value
        _consumer.last_stable_offset.assert_called_once_with(TP1)

    def test_topic_partitions(self, *, cthread, _consumer):
        cthread._consumer = None
        assert cthread.topic_partitions('foo') is None
        cthread._consumer = _consumer
        assert cthread.topic_partitions('foo') is (
            _consumer._coordinator._metadata_snapshot.get.return_value)

    @pytest.mark.asyncio
    async def test_earliest_offsets(self, *, cthread, _consumer):
        with self.assert_calls_thread(
                cthread, _consumer, _consumer.beginning_offsets, (TP1,)):
            await cthread.earliest_offsets(TP1)

    @pytest.mark.asyncio
    async def test_highwaters(self, *, cthread, _consumer):
        with self.assert_calls_thread(
                cthread, _consumer, cthread._highwaters, (TP1,)):
            await cthread.highwaters(TP1)

    @pytest.mark.asyncio
    async def test__highwaters(self, *, cthread, _consumer):
        cthread.consumer.in_transaction = False
        cthread._consumer = _consumer
        assert await cthread._highwaters([TP1]) is (
            _consumer.end_offsets.coro.return_value)

    @pytest.mark.asyncio
    async def test__highwaters__in_transaction(self, *, cthread, _consumer):
        cthread.consumer.in_transaction = True
        cthread._consumer = _consumer
        assert await cthread._highwaters([TP1]) == {
            TP1: _consumer.last_stable_offset.return_value,
        }

    def test__ensure_consumer(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        assert cthread._ensure_consumer() is _consumer
        cthread._consumer = None
        with pytest.raises(ConsumerNotStarted):
            cthread._ensure_consumer()

    @pytest.mark.asyncio
    async def test_getmany(self, *, cthread, _consumer):
        timeout = 13.1
        active_partitions = {TP1}
        with self.assert_calls_thread(
                cthread, _consumer, cthread._fetch_records,
                _consumer,
                active_partitions,
                timeout=timeout,
                max_records=_consumer._max_poll_records):
            await cthread.getmany(active_partitions, timeout)

    @pytest.mark.asyncio
    async def test__fetch_records__flow_inactive(self, *, cthread, _consumer):
        cthread.consumer.flow_active = False
        assert await cthread._fetch_records(_consumer, set()) == {}

    @pytest.mark.asyncio
    async def test__fetch_records_consumer_closed(self, *, cthread, _consumer):
        _consumer._closed = True
        with pytest.raises(ConsumerStoppedError):
            await cthread._fetch_records(_consumer, set())

    @pytest.mark.asyncio
    async def test__fetch_records_fetcher_closed(self, *, cthread, _consumer):
        _consumer._closed = False
        _consumer._fetcher._closed = True
        with pytest.raises(ConsumerStoppedError):
            await cthread._fetch_records(_consumer, set())

    @pytest.mark.asyncio
    async def test__fetch_records(self, *, cthread, _consumer):
        _consumer._closed = False
        fetcher = _consumer._fetcher
        fetcher._closed = False
        fetcher._subscriptions.fetch_context = MagicMock()
        fetcher.fetched_records = AsyncMock()
        ret = await cthread._fetch_records(
            _consumer, {TP1}, timeout=312.3, max_records=1000)
        assert ret is fetcher.fetched_records.coro.return_value
        fetcher.fetched_records.assert_called_once_with(
            {TP1},
            timeout=312.3,
            max_records=1000,
        )

    @pytest.mark.asyncio
    async def test_create_topic(self, *, cthread, _consumer):
        topic = 'topic'
        partitions = 30
        replication = 11
        config = {'cfg': 'foo'}
        timeout = 30.312
        retention = 1210320.3
        compacting = True
        deleting = False
        ensure_created = True

        transport = cthread.transport
        with self.assert_calls_thread(
                cthread, _consumer, transport._create_topic,
                cthread,
                _consumer._client,
                topic,
                partitions,
                replication,
                config=config,
                timeout=int(timeout * 1000.0),
                retention=int(retention * 1000.0),
                compacting=compacting,
                deleting=deleting,
                ensure_created=ensure_created):
            await cthread.create_topic(
                topic,
                partitions,
                replication,
                config=config,
                timeout=timeout,
                retention=retention,
                compacting=compacting,
                deleting=deleting,
                ensure_created=ensure_created,
            )

    @pytest.mark.asyncio
    async def test_create_topic_invalid_name(self, cthread, _consumer):
        topic = ''.join(random.choices(
            string.ascii_uppercase, k=TOPIC_LENGTH_MAX + 1))
        cthread._consumer = _consumer

        msg = f'Topic name {topic!r} is too long (max={TOPIC_LENGTH_MAX})'
        with pytest.raises(ValueError):
            await cthread.create_topic(topic, 1, 1)
            pytest.fail(msg)

    def test_key_partition(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread._partitioner = Mock(name='partitioner')
        metadata = _consumer._client.cluster
        metadata.partitions_for_topic.return_value = [1, 2, 3]
        metadata.available_partitions_for_topic.return_value = [2, 3]

        cthread.key_partition('topic', 'k', None)
        cthread._partitioner.assert_called_once_with(
            'k', [1, 2, 3], [2, 3],
        )

        with pytest.raises(AssertionError):
            cthread.key_partition('topic', 'k', -1)
        with pytest.raises(AssertionError):
            cthread.key_partition('topic', 'k', 4)

        assert cthread.key_partition('topic', 'k', 3) == 3

    def test_key_partition__no_metadata(self, *, cthread, _consumer):
        cthread._consumer = _consumer
        cthread._partitioner = Mock(name='partitioner')
        metadata = _consumer._client.cluster
        metadata.partitions_for_topic.return_value = None

        assert cthread.key_partition('topic', 'k', None) is None

    @contextmanager
    def assert_calls_thread(self, cthread, _consumer, method, *args, **kwargs):
        cthread._consumer = _consumer
        cthread.call_thread = AsyncMock()
        try:
            yield
        finally:
            cthread.call_thread.assert_called_once_with(
                method, *args, **kwargs)


class MyPartitioner:
    ...


my_partitioner = MyPartitioner()


class test_Producer:

    @pytest.fixture()
    def producer(self, *, app, _producer):
        producer = Producer(app.transport)
        producer._producer = _producer
        return producer

    @pytest.fixture()
    def _producer(self):
        return Mock(
            name='AIOKafkaProducer',
            autospec=aiokafka.AIOKafkaProducer,
            start=AsyncMock(),
            stop=AsyncMock(),
            begin_transaction=AsyncMock(),
            commit_transaction=AsyncMock(),
            abort_transaction=AsyncMock(),
            stop_transaction=AsyncMock(),
            maybe_begin_transaction=AsyncMock(),
            commit=AsyncMock(),
            send=AsyncMock(),
            flush=AsyncMock(),
        )

    @pytest.mark.conf(producer_partitioner=my_partitioner)
    def test_producer__uses_custom_partitioner(self, *, producer):
        assert producer.partitioner is my_partitioner

    @pytest.mark.asyncio
    async def test_begin_transaction(self, *, producer, _producer):
        await producer.begin_transaction('tid')
        _producer.begin_transaction.assert_called_once_with('tid')

    @pytest.mark.asyncio
    async def test_commit_transaction(self, *, producer, _producer):
        await producer.commit_transaction('tid')
        _producer.commit_transaction.assert_called_once_with('tid')

    @pytest.mark.asyncio
    async def test_abort_transaction(self, *, producer, _producer):
        await producer.abort_transaction('tid')
        _producer.abort_transaction.assert_called_once_with('tid')

    @pytest.mark.asyncio
    async def test_stop_transaction(self, *, producer, _producer):
        await producer.stop_transaction('tid')
        _producer.stop_transaction.assert_called_once_with('tid')

    @pytest.mark.asyncio
    async def test_maybe_begin_transaction(self, *, producer, _producer):
        await producer.maybe_begin_transaction('tid')
        _producer.maybe_begin_transaction.assert_called_once_with('tid')

    @pytest.mark.asyncio
    async def test_commit_transactions(self, *, producer, _producer):
        tid_to_offset_map = {'t1': {TP1: 1001}, 't2': {TP2: 2002}}
        await producer.commit_transactions(
            tid_to_offset_map, 'group_id', start_new_transaction=False)
        _producer.commit.assert_called_once_with(
            tid_to_offset_map, 'group_id', start_new_transaction=False)

    def test__settings_extra(self, *, producer, app):
        app.in_transaction = True
        assert producer._settings_extra() == {'acks': 'all'}
        app.in_transaction = False
        assert producer._settings_extra() == {}

    def test__new_producer(self, *, producer):
        self.assert_new_producer(producer)

    @pytest.mark.parametrize('expected_args', [
        pytest.param({'api_version': '0.10'},
                     marks=pytest.mark.conf(
                         producer_api_version='0.10')),
        pytest.param({'acks': 'all'},
                     marks=pytest.mark.conf(
                         producer_acks='all')),
        pytest.param({'bootstrap_servers': ['a:9092', 'b:9092']},
                     marks=pytest.mark.conf(
                         broker='kafka://a:9092;b:9092')),
        pytest.param({'client_id': 'foo'},
                     marks=pytest.mark.conf(
                         broker_client_id='foo')),
        pytest.param({'compression_type': 'snappy'},
                     marks=pytest.mark.conf(
                         producer_compression_type='snappy')),
        pytest.param({'linger_ms': 9345},
                     marks=pytest.mark.conf(
                         producer_linger_ms=9345)),
        pytest.param({'max_batch_size': 41223},
                     marks=pytest.mark.conf(
                         producer_max_batch_size=41223)),
        pytest.param({'max_request_size': 183831},
                     marks=pytest.mark.conf(
                         producer_max_request_size=183831)),
        pytest.param({'request_timeout_ms': 1234134000},
                     marks=pytest.mark.conf(
                         producer_request_timeout=1234134)),
        pytest.param(
            {'security_protocol': 'SASL_PLAINTEXT',
             'sasl_mechanism': 'PLAIN',
             'sasl_plain_username': 'uname',
             'sasl_plain_password': 'pw',
             'ssl_context': None},
            marks=pytest.mark.conf(
                broker_credentials=auth.SASLCredentials(
                    username='uname',
                    password='pw',
                    mechanism='PLAIN'),
            ),
        ),
    ])
    def test__new_producer__using_settings(self, expected_args, *,
                                           app, producer):
        self.assert_new_producer(producer, **expected_args)

    def assert_new_producer(self, producer,
                            acks=-1,
                            api_version='auto',
                            bootstrap_servers=['localhost:9092'],  # noqa,
                            client_id=f'faust-{faust.__version__}',
                            compression_type=None,
                            linger_ms=0,
                            max_batch_size=16384,
                            max_request_size=1000000,
                            request_timeout_ms=1200000,
                            security_protocol='PLAINTEXT',
                            **kwargs):
        with patch('aiokafka.AIOKafkaProducer') as AIOKafkaProducer:
            p = producer._new_producer()
            assert p is AIOKafkaProducer.return_value
            AIOKafkaProducer.assert_called_once_with(
                acks=acks,
                api_version=api_version,
                bootstrap_servers=bootstrap_servers,
                client_id=client_id,
                compression_type=compression_type,
                linger_ms=linger_ms,
                max_batch_size=max_batch_size,
                max_request_size=max_request_size,
                request_timeout_ms=request_timeout_ms,
                security_protocol=security_protocol,
                loop=producer.loop,
                partitioner=producer.partitioner,
                **kwargs,
            )

    def test__new_producer__default(self, *, producer):
        p = producer._new_producer()
        assert isinstance(p, aiokafka.AIOKafkaProducer)

    def test__new_producer__in_transaction(self, *, producer):
        producer.app.in_transaction = True
        p = producer._new_producer()
        assert isinstance(p, aiokafka.MultiTXNProducer)

    def test__producer_type(self, *, producer, app):
        app.in_transaction = True
        assert producer._producer_type is aiokafka.MultiTXNProducer
        app.in_transaction = False
        assert producer._producer_type is aiokafka.AIOKafkaProducer

    @pytest.mark.asyncio
    async def test_create_topic(self, *, producer, _producer):
        _producer.client = Mock(
            force_metadata_update=AsyncMock(),
        )
        producer.transport = Mock(
            _create_topic=AsyncMock(),
        )
        await producer.create_topic(
            'foo', 100, 3,
            config={'x': 'y'},
            timeout=30.3,
            retention=300.3,
            compacting=True,
            deleting=True,
            ensure_created=True,
        )
        producer.transport._create_topic.coro.assert_called_once_with(
            producer,
            _producer.client,
            'foo',
            100,
            3,
            config={'x': 'y'},
            timeout=int(30.3 * 1000.0),
            retention=int(300.3 * 1000.0),
            compacting=True,
            deleting=True,
            ensure_created=True,
        )

    def test__ensure_producer(self, *, producer, _producer):
        assert producer._ensure_producer() is _producer
        producer._producer = None
        with pytest.raises(NotReady):
            producer._ensure_producer()

    @pytest.mark.asyncio
    async def test_on_start(self, *, producer, loop):
        producer._new_producer = Mock(
            name='_new_producer',
            return_value=Mock(
                start=AsyncMock(),
            ),
        )
        _producer = producer._new_producer.return_value
        producer.beacon = Mock()

        await producer.on_start()
        assert producer._producer is _producer
        producer._new_producer.assert_called_once_with()
        producer.beacon.add.assert_called_with(_producer)
        _producer.start.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_stop(self, *, producer, _producer):
        await producer.on_stop()
        assert producer._producer is None
        _producer.stop.assert_called_once_with()

    def test_supports_headers__not_ready(self, *, producer):
        producer._producer.client = None
        with pytest.raises(NotReady):
            producer.supports_headers()

    @pytest.mark.asyncio
    async def test_send(self, producer, _producer):
        await producer.send(
            'topic', 'k', 'v', 3, 100, {'foo': 'bar'},
            transactional_id='tid',
        )
        _producer.send.assert_called_once_with(
            'topic', 'v',
            key='k',
            partition=3,
            timestamp_ms=100 * 1000.0,
            headers=[('foo', 'bar')],
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    @pytest.mark.conf(producer_api_version='0.10')
    async def test_send__request_no_headers(self, producer, _producer):
        await producer.send(
            'topic', 'k', 'v', 3, 100, {'foo': 'bar'},
            transactional_id='tid',
        )
        _producer.send.assert_called_once_with(
            'topic', 'v',
            key='k',
            partition=3,
            timestamp_ms=100 * 1000.0,
            headers=None,
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    @pytest.mark.conf(producer_api_version='0.11')
    async def test_send__kafka011_supports_headers(self, producer, _producer):
        await producer.send(
            'topic', 'k', 'v', 3, 100, {'foo': 'bar'},
            transactional_id='tid',
        )
        _producer.send.assert_called_once_with(
            'topic', 'v',
            key='k',
            partition=3,
            timestamp_ms=100 * 1000.0,
            headers=[('foo', 'bar')],
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    @pytest.mark.conf(producer_api_version='auto')
    async def test_send__auto_passes_headers(self, producer, _producer):
        await producer.send(
            'topic', 'k', 'v', 3, 100, [('foo', 'bar')],
            transactional_id='tid',
        )
        _producer.send.assert_called_once_with(
            'topic', 'v',
            key='k',
            partition=3,
            timestamp_ms=100 * 1000.0,
            headers=[('foo', 'bar')],
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    async def test_send__no_headers(self, producer, _producer):
        await producer.send(
            'topic', 'k', 'v', 3, 100, None,
            transactional_id='tid',
        )
        _producer.send.assert_called_once_with(
            'topic', 'v',
            key='k',
            partition=3,
            timestamp_ms=100 * 1000.0,
            headers=None,
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    async def test_send__no_timestamp(self, producer, _producer):
        await producer.send(
            'topic', 'k', 'v', 3, None, None,
            transactional_id='tid',
        )
        _producer.send.assert_called_once_with(
            'topic', 'v',
            key='k',
            partition=3,
            timestamp_ms=None,
            headers=None,
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    async def test_send__KafkaError(self, producer, _producer):
        _producer.send.coro.side_effect = KafkaError()
        with pytest.raises(ProducerSendError):
            await producer.send(
                'topic', 'k', 'v', 3, None, None,
                transactional_id='tid',
            )

    @pytest.mark.asyncio
    async def test_send_and_wait(self, producer):
        producer.send = AsyncMock(return_value=done_future(done_future()))

        await producer.send_and_wait(
            'topic', 'k', 'v', 3, 100, [('a', 'b')],
            transactional_id='tid')
        producer.send.assert_called_once_with(
            'topic',
            key='k',
            value='v',
            partition=3,
            timestamp=100,
            headers=[('a', 'b')],
            transactional_id='tid',
        )

    @pytest.mark.asyncio
    async def test_flush(self, *, producer, _producer):
        producer._producer = None
        await producer.flush()
        producer._producer = _producer
        await producer.flush()
        _producer.flush.assert_called_once_with()

    def test_key_partition(self, *, producer, _producer):
        x = producer.key_partition('topic', 'k')
        assert x == TP('topic', _producer._partition.return_value)

    def test_supports_headers(self, *, producer):
        producer._producer.client.api_version = (0, 11)
        assert producer.supports_headers()


class test_Transport:

    @pytest.fixture()
    def transport(self, *, app):
        return Transport(url=['aiokafka://'], app=app)

    def test_constructor(self, *, transport):
        assert transport._topic_waiters == {}

    def test__topic_config(self, *, transport):
        assert transport._topic_config() == {}

    def test__topic_config__retention(self, *, transport):
        assert transport._topic_config(retention=3000.3) == {
            'retention.ms': 3000.3,
        }

    def test__topic_config__compacting(self, *, transport):
        assert transport._topic_config(compacting=True) == {
            'cleanup.policy': 'compact',
        }

    def test__topic_config__deleting(self, *, transport):
        assert transport._topic_config(deleting=True) == {
            'cleanup.policy': 'delete',
        }

    def test__topic_config__combined(self, *, transport):
        res = transport._topic_config(
            compacting=True, deleting=True, retention=3000.3)
        assert res == {
            'retention.ms': 3000.3,
            'cleanup.policy': 'compact,delete',
        }

    @pytest.mark.asyncio
    async def test__create_topic(self, *, transport):
        client = Mock(name='client')
        transport._topic_waiters['foo'] = AsyncMock()
        await transport._create_topic(
            transport,
            client,
            topic='foo',
            partitions=100,
            replication=3,
        )
        transport._topic_waiters['foo'].coro.assert_called_with()

    @pytest.mark.asyncio
    async def test__create_topic__missing(self, *, transport, loop):
        client = Mock(name='client')
        transport._topic_waiters.clear()
        with patch('faust.transport.drivers.aiokafka.StampedeWrapper') as SW:
            SW.return_value = AsyncMock()
            await transport._create_topic(
                transport,
                client,
                topic='foo',
                partitions=100,
                replication=3,
            )
            SW.assert_called_once_with(
                transport._really_create_topic,
                transport,
                client,
                'foo',
                100,
                3,
                loop=loop,
            )
            SW.return_value.coro.assert_called_once_with()
            assert transport._topic_waiters['foo'] is SW.return_value

    @pytest.mark.asyncio
    async def test__create_topic__raises(self, *, transport, loop):
        client = Mock(name='client')
        transport._topic_waiters.clear()
        with patch('faust.transport.drivers.aiokafka.StampedeWrapper') as SW:
            SW.return_value = AsyncMock()
            SW.return_value.coro.side_effect = KeyError('foo')
            with pytest.raises(KeyError):
                await transport._create_topic(
                    transport,
                    client,
                    topic='foo',
                    partitions=100,
                    replication=3,
                )
            assert 'foo' not in transport._topic_waiters


@pytest.mark.parametrize('credentials,ssl_context,expected', [
    (None, {}, {
        'security_protocol': 'SSL', 'ssl_context': {},
    }),
    (None, None, {
        'security_protocol': 'PLAINTEXT',
    }),
    (auth.SSLCredentials({}), None, {
        'security_protocol': 'SSL', 'ssl_context': {},
    }),
    (auth.SASLCredentials(username='foo', password='bar'), None, {
        'security_protocol': 'SASL_PLAINTEXT',
        'sasl_mechanism': 'PLAIN',
        'sasl_plain_username': 'foo',
        'sasl_plain_password': 'bar',
        'ssl_context': None,
    }),
    (auth.GSSAPICredentials(kerberos_service_name='service',
                            kerberos_domain_name='moo'), None, {
        'security_protocol': 'SASL_PLAINTEXT',
        'sasl_mechanism': 'GSSAPI',
        'sasl_kerberos_service_name': 'service',
        'sasl_kerberos_domain_name': 'moo',
        'ssl_context': None,
    }),
])
def test_credentials_to_aiokafka(credentials, ssl_context, expected):
    assert credentials_to_aiokafka_auth(credentials, ssl_context) == expected


def test_credentials_to_aiokafka__invalid():
    with pytest.raises(ImproperlyConfigured):
        credentials_to_aiokafka_auth(object())



import pytest
from faust.agents import current_agent


@pytest.mark.asyncio
async def test_sets_current_agent(*, app, event_loop):
    assert current_agent() is None

    agent_body_executed = event_loop.create_future()

    @app.agent()
    async def agent(stream):
        agent_body_executed.set_result(True)
        assert current_agent() is agent
        async for item in stream:
            ...

    await agent.start()
    await agent_body_executed
    await agent.stop()



import asyncio
import os
import random
import sys
from itertools import count
from time import monotonic
from faust.cli import option

__all__ = ['Benchmark']

TIME_EVERY = 10_000
BENCH_TYPE = os.environ.get('F_BENCH', 'worker')
ACKS = int(os.environ.get('F_ACKS') or 0)
BSIZE = int(os.environ.get('F_BSIZE', 16384))
LINGER = int(os.environ.get('F_LINGER', 0))


class Benchmark:
    _agent = None
    _produce = None

    produce_options = [
        option('--max-latency',
               type=float, default=0.0, envvar='PRODUCE_LATENCY',
               help='Add delay of (at most) n seconds between publishing.'),
        option('--max-messages',
               type=int, default=None,
               help='Send at most N messages or 0 for infinity.'),
    ]

    def __init__(self, app, topic, *,
                 n=TIME_EVERY,
                 consume_topic=None):
        self.app = app
        self.topic = topic
        if consume_topic is None:
            consume_topic = topic
        self.consume_topic = consume_topic
        self.n = n

        self.app.finalize()
        self.app.conf.producer_acks = ACKS
        self.app.conf.producer_max_batch_size = BSIZE
        self.app.conf.producer_linger_ms = LINGER

    def install(self, main_name):
        self.create_benchmark_agent()
        self.create_produce_command()
        if main_name == '__main__':
            # we use this for cProfile
            bench_args = {
                'worker': ['worker', '-l', 'info'],
                'produce': ['produce'],
            }
            if len(sys.argv) < 2:
                sys.argv.extend(bench_args[BENCH_TYPE])
            self.app.main()

    def create_benchmark_agent(self):
        self._agent = self.app.agent(self.consume_topic)(self.process)

    async def process(self, stream):
        time_last = None
        async for i, value in stream.enumerate():
            if not i:
                time_last = monotonic()
            await self.process_value(value)
            if i and not i % self.n:  # noqa: S001
                now = monotonic()
                runtime, time_last = now - time_last, now
                print(f'RECV {i} in {runtime}s')

    async def process_value(self, value):
        # put worker benchmark-specific code in here
        ...

    def create_produce_command(self):
        self._produce = self.app.command(*self.produce_options)(self.produce)

    async def produce(self, max_latency: float, max_messages: int, **kwargs):
        i = 0
        time_start = None
        app = self.app
        topic = self.topic
        for i, (key, value) in enumerate(self.generate_values(max_messages)):
            callback = None
            if not i:
                time_start = monotonic()
                time_1st = monotonic()

                def on_published(meta):
                    print(f'1ST OK: {meta} AFTER {monotonic() - time_1st}s')
                callback = on_published
            await topic.send(key=key, value=value, callback=callback)
            if i and not i % self.n:  # noqa: S001
                print(f'+SEND {i} in {monotonic() - time_start}s')
                time_start = monotonic()
            if max_latency:
                await asyncio.sleep(random.uniform(0, max_latency))
        await asyncio.sleep(10)
        await app.producer.stop()
        print(f'Time spent on {i} messages: {monotonic() - time_start}s')

    def generate_values(self, max_messages: int):
        # override this to generate other types of values
        return (
            # convert int to bytes, cannot use bytes(int) as that gives
            # character code in Python3.
            (None, str(i).encode())
            for i in (range(max_messages) if max_messages else count())
        )



import ssl
import pytest
from faust.auth import GSSAPICredentials, SASLCredentials, SSLCredentials
from faust.types.auth import AuthProtocol, SASLMechanism
from mode.utils.mocks import Mock, patch


class test_SASLCredentials:

    @pytest.mark.parametrize('reason,credentials,expected_fields', [
        pytest.param(
            'retains arguments',
            SASLCredentials(username='george', password='pw1'),
            {'username': 'george',
             'password': 'pw1',
             'mechanism': SASLMechanism.PLAIN,
             'protocol': AuthProtocol.SASL_PLAINTEXT}),
        pytest.param(
            'mechanism from str',
            SASLCredentials(username='george',
                            password='pw1',
                            mechanism='GSSAPI'),
            {'mechanism': SASLMechanism.GSSAPI}),
        pytest.param(
            'mechanism from enum',
            SASLCredentials(username='george',
                            password='pw1',
                            mechanism=SASLMechanism.GSSAPI),
            {'mechanism': SASLMechanism.GSSAPI}),
        pytest.param(
            'ssl context gives SASL_SSL',
            SASLCredentials(username='george',
                            password='pw1',
                            ssl_context={'xxx': 'yyy'}),
            {'username': 'george',
             'password': 'pw1',
             'ssl_context': {'xxx': 'yyy'},
             'protocol': AuthProtocol.SASL_SSL}),
    ])
    def test_constructor(self, credentials, expected_fields, reason):
        assert repr(credentials)
        for field, value in expected_fields.items():
            assert getattr(credentials, field) == value, reason


class test_GSSAPICredentials:

    @pytest.mark.parametrize('reason,credentials,expected_fields', [
        pytest.param(
            'retains arguments',
            GSSAPICredentials(kerberos_service_name='george',
                              kerberos_domain_name='domain'),
            {'kerberos_service_name': 'george',
             'kerberos_domain_name': 'domain',
             'mechanism': SASLMechanism.GSSAPI,
             'protocol': AuthProtocol.SASL_PLAINTEXT}),
        pytest.param(
            'mechanism given as str',
            GSSAPICredentials(kerberos_service_name='george',
                              kerberos_domain_name='domain',
                              mechanism='PLAIN'),
            {'mechanism': SASLMechanism.PLAIN}),
        pytest.param(
            'mechanism given as enum',
            GSSAPICredentials(kerberos_service_name='george',
                              kerberos_domain_name='domain',
                              mechanism=SASLMechanism.PLAIN),
            {'mechanism': SASLMechanism.PLAIN}),
        pytest.param(
            'ssl context gives SASL_SSL',
            GSSAPICredentials(kerberos_service_name='george',
                              kerberos_domain_name='domain',
                              ssl_context={'xxx': 'yyy'}),
            {'kerberos_service_name': 'george',
             'kerberos_domain_name': 'domain',
             'ssl_context': {'xxx': 'yyy'},
             'protocol': AuthProtocol.SASL_SSL}),
    ])
    def test_constructor(self, credentials, expected_fields, reason):
        assert repr(credentials)
        for field, value in expected_fields.items():
            assert getattr(credentials, field) == value, reason


class test_SSLCredentials:

    def test_constructor(self):
        with patch('faust.auth.ssl.create_default_context') as cdc:
            c = SSLCredentials(
                purpose=ssl.Purpose.SERVER_AUTH,
                cafile='/foo/bar/ca.file',
                capath='/foo/bar/',
                cadata='moo',
            )
            assert c.context is cdc.return_value
            assert repr(c)
            cdc.assert_called_once_with(
                purpose=ssl.Purpose.SERVER_AUTH,
                cafile='/foo/bar/ca.file',
                capath='/foo/bar/',
                cadata='moo',
            )

    def test_having_context(self):
        context = Mock(name='context')
        c = SSLCredentials(context)
        assert c.context is context



def test_json(faust_json):
    exitcode, models, stderr = faust_json('models', '--builtins')
    assert not exitcode

    assert {'name': 'app.Arena', 'help': '<N/A>'} in models
    assert {'name': 'app.Point', 'help': '<N/A>'} in models
    assert {'name': '@ReqRepResponse',
            'help': 'Request-Reply response.'} in models

    names = [model['name'] for model in models]
    assert (names.index('@ReqRepResponse') <
            names.index('app.Arena') <
            names.index('app.Point'))  # sorted


def test_tabulated(faust):
    exitcode, stdout, stderr = faust('models', '--builtins')
    assert not exitcode
    assert b'Arena' in stdout


def test_colors(faust_color):
    exitcode, stdout, stderr = faust_color('models', '--builtins')
    assert not exitcode
    assert b'Point' in stdout


def test_json_no_local(faust_json):
    exitcode, models, stderr = faust_json('models')
    assert not exitcode

    names = [model['name'] for model in models]
    assert 'app.Arena' in names
    assert 'app.Point' in names
    assert names.index('app.Arena') < names.index('app.Point')  # sorted
